#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""grid_definition.py: Contains the grid definition class used with pygeostat DataFiles"""

#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import numpy as np


class GridDef:

    """
    Class containing GSLIB grid definition.

    Given either a GSLIB style grid string or an array arranged as [nx, xmn, xsiz, ny, ymn, ysiz,
    nz, zmn, zsiz] initialize the GridDef Class

    """

    def __init__(self, griddef=None, grid_str=None, grid_arr=None, grid_file=None):

        if isinstance(griddef, str) and grid_str is None:
            grid_str = griddef
        elif isinstance(griddef, list) and grid_arr is None:
            grid_arr = griddef

        if grid_file is not None:
            if grid_str is not None:
                print('WARNING: Both a string and an file supplied to grid initialization')
                print('  Assuming the file')
            with open(grid_file) as f:
                grid_str = f.read()

        if grid_str is not None:
            temp_grid_arr = self._parse_grid_string(grid_str)
            if grid_arr is not None:
                if grid_arr != temp_grid_arr:
                    print('WARNING: Both a string and an array supplied to grid initialization')
                    print('  Assuming that the array should be used')
                    temp_grid_arr = grid_arr
        elif grid_arr is not None:
            temp_grid_arr = grid_arr
            if grid_file is not None:
                print('WARNING: Both a string and an file supplied to grid initialization')
                print('  Assuming the array')
        else:
            print('WARNING: No grid definition supplied')
            print('  Assuming zeroes')
            temp_grid_arr = [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        self.nx = int(temp_grid_arr[0])
        self.xmn = float(temp_grid_arr[1])
        self.xsiz = float(temp_grid_arr[2])
        self.ny = int(temp_grid_arr[3])
        self.ymn = float(temp_grid_arr[4])
        self.ysiz = float(temp_grid_arr[5])
        self.nz = int(temp_grid_arr[6])
        self.zmn = float(temp_grid_arr[7])
        self.zsiz = float(temp_grid_arr[8])
        self.xlimits = self.extents()[0]
        self.ylimits = self.extents()[1]
        self.zlimits = self.extents()[2]
        self.xlength = (self.xlimits[1] - self.xlimits[0])
        self.ylength = (self.ylimits[1] - self.ylimits[0])
        self.zlength = (self.zlimits[1] - self.zlimits[0])

    def __str__(self):
        """Return the string representation of the current grid definition"""
        return "{} {} {} \n{} {} {} \n{} {} {}".format(self.nx, self.xmn, self.xsiz, self.ny,
                                                       self.ymn, self.ysiz, self.nz, self.zmn,
                                                       self.zsiz)

    def __repr__(self):
        " pretty printing for jupyter notebook workflows"
        return "Pygeostat GridDef:\n{}".format(self)

    def copy(self):
        """ return a copy of this object """
        return GridDef(grid_str=str(self))

    def _parse_grid_string(self, grid_str):
        """Parse the GSLIB style grid string into a list of floats"""
        arr = []
        for line in str.splitlines(grid_str):
            # Check that we have data
            if len(line) > 0:
                # Append the first 3 entries
                arr.append(line.split()[0])
                arr.append(line.split()[1])
                arr.append(line.split()[2])
        assert len(arr) == 9
        return arr

    def count(self):
        """Return the number of blocks in the current grid definition"""
        return self.nx * self.ny * self.nz

    def origin(self):
        """Return the origin of the current grid definition"""
        [(xmin, xmax), (ymin, ymax), (zmin, zmax)] = self.extents()
        return (xmin, ymin, zmin)


    def _get_grid_array(self):
        """ Return the array of grid parameters (nx, xmn, xsiz, ny, ymn, ysiz, nz, zmn, zsiz)  """
        return (self.nx, self.xmn, self.xsiz, self.ny, self.ymn, self.ysiz, self.nz, self.zmn,
                self.zsiz)
    grid_array = property(_get_grid_array)


    def _get_block_volume(self):
        """Return the volume of one block in the current grid definition"""
        return self.xsiz * self.ysiz * self.zsiz

    block_volume = property(_get_block_volume)

    def convert_to_2d(self, orient='xy'):
        """
        Flattens a grid to 2D by default on the xy plane, returning a new 2D GridDef object
        """
        if orient == 'xy':
            return GridDef('%i %f %f\n%i %f %f\n1 0.5 1' % (self.nx, self.xmn, self.xsiz,
                                                            self.ny, self.ymn, self.ysiz))
        elif orient == 'xz':
            return GridDef('%i %f %f\n%i %f %f\n1 0.5 1' % (self.nx, self.xmn, self.xsiz,
                                                            self.nz, self.zmn, self.zsiz))
        elif orient == 'yz':
            return GridDef('%i %f %f\n%i %f %f\n1 0.5 1' % (self.ny, self.ymn, self.ysiz,
                                                            self.nz, self.zmn, self.zsiz))

    def extents(self, orient=None):
        """
        Return the extents of the current grid definition.

        Parameters:
            orient(str): acceptable is 'x','y', or 'z' to give the dimensions along that direction

        Returns:
            various tuples based on what was passed

        """
        xmin = self.xmn - self.xsiz / 2.0
        xmax = xmin + (self.nx) * self.xsiz
        ymin = self.ymn - self.ysiz / 2.0
        ymax = ymin + (self.ny) * self.ysiz
        zmin = self.zmn - self.zsiz / 2.0
        zmax = zmin + (self.nz) * self.zsiz
        if orient is None:
            return [(xmin, xmax), (ymin, ymax), (zmin, zmax)]
        elif orient == 'x':
            return (xmin, xmax)
        elif orient == 'y':
            return (ymin, ymax)
        elif orient == 'z':
            return (zmin, zmax)
        else:
            raise ValueError('Unacceptable orient was provided. x, y and z are the accpetable options')

    def outline_points(self, orient='xy'):
        """
        return the xpts and ypts for plotline an outline of the current grid
        definition in the defined orientation

        Parameters:
            orient(str): The orientation to return the corner points of the grid
                 ``'xy'``, ``'xz'``, ``'yz'`` are the only accepted values

        Returns:
            xpts, ypts (float, float): the corner points of the grid in the specified orientation

        """

        if orient.lower() == 'xy':
            xpts = [self.xlimits[0], self.xlimits[0], self.xlimits[1],
                    self.xlimits[1], self.xlimits[0]]
            ypts = [self.ylimits[0], self.ylimits[1], self.ylimits[1],
                    self.ylimits[0], self.ylimits[0]]
        elif orient.lower() == 'xz':
            xpts = [self.xlimits[0], self.xlimits[0], self.xlimits[1],
                    self.xlimits[1], self.xlimits[0]]
            ypts = [self.zlimits[0], self.zlimits[1], self.zlimits[1],
                    self.zlimits[0], self.zlimits[0]]
        elif orient.lower() == 'yz':
            xpts = [self.ylimits[0], self.ylimits[0], self.ylimits[1],
                    self.ylimits[1], self.ylimits[0]]
            ypts = [self.zlimits[0], self.zlimits[1], self.zlimits[1],
                    self.zlimits[0], self.zlimits[0]]
        else:
            raise ValueError('''`xy`, `xz`, `yz` are the only accepted values for orient''')

        return xpts, ypts

    def index3d_to_index1d(self, ix, iy, iz):
        """
        Will return the 0-indexed 1-Dimensional grid index given the 3 dimensional indices
        as well as a True/False indicator for inside or outside the grid definition.

        Parameters:
            ix (int or numpy.ndarray): x index or n-length array of x indices
            iy (int or numpy.ndarray): y index or n-length array of y indices
            iz (int or numpy.ndarray): z index or n-length array of z indices

        Returns:
            idx (int or numpy.ndarray): 1-d index or n-length array of 1-d indices
            ingrid (bool or numpy.ndarray): in (True) or out (False) of grid, returned if
            ingrid=True

        Examples:

            Calculate a 1d grid index based on a 3d index (integers). Returns the index as an
            integer, as well as a boolean of whether the index is in the grid:

            >>> ix, iy, iz = 2, 4, 0
            >>> idx, ingrid = griddef.index3d_to_index1d(ix, iy, iz)

            Calculate 1d grid indices based on 3d indices (numpy arrays). Returns an array
            of indices, as well as a boolean array of whether the index is in the grid:

            >>> ix, iy, iz = np.arange(0, 5), np.arange(0, 5), np.zeros(5)
            >>> idx, ingrid = griddef.index3d_to_index1d(ix, iy, iz)

        """
        def isall(objtype, objects): return all(isinstance(ob, objtype) for ob in objects)
        if isall(int, [ix, iy, iz]):
            idx = int(ix + iy * self.nx + iz * self.nx * self.ny)
            if idx < self.count() and idx >= 0:
                ingrid = True
            else:
                ingrid = False
        elif isall(np.ndarray, [ix, iy, iz]):
            # Ensure the past indices are integers
            ix = ix.astype(int, casting='safe')
            iy = ix.astype(int, casting='safe')
            iz = ix.astype(int, casting='safe')
            idx = np.add(ix, iy * self.nx, iz * self.nx * self.ny, dtype=int)
            ingrid = np.logical_and(idx >= 0, idx < self.count())
        else:
            raise TypeError(('ix, iy and iz should be the same type, which is either'
                             'int or numpy.ndarray of matching length'))
        return idx, ingrid

    def index1d_to_index3d(self, idx):
        """Will return the 3-dimensional indices given a 1 dimensional index as well as
        a True/False indicator for inside or outside the grid definition.

        Parameters:
            idx (int or np.ndarray): 1 dimensional index or numpy array of indices

        Returns:
            ix (int or np.ndarray): x index or numpy array of x indices
            iy (int or np.ndarray): y index or numpy array of y indices
            iz (int or np.ndarray): z index or numpy array of z indices
            ingrid (bool or np.ndarray): In (True) or Out (False) of grid

        Examples:

            Calculate a 3d grid index based on a 1d index (integers). Returns the 3d index as an
            integer, as well as a boolean of whether the index is in the grid:

            >>> idx = 918
            >>> ix, iy, iz, ingrid = griddef.index1d_to_index3d(idx)

            Calculate 3d grid indices based on 1d indices (numpy array). Returns a arrays
            of 3d indices, as well as a boolean array of whether the indices are in the grid:

            >>> idx = np.array([0, 230, 460, 690, 920])
            >>> ix, iy, iz, ingrid = griddef.index1d_to_index3d(idx)

        """
        if isinstance(idx, (int, np.integer)):
            ix = int((idx % (self.nx * self.ny)) % self.nx)
            iy = int((idx % (self.nx * self.ny)) / self.nx)
            iz = int((idx / (self.nx * self.ny)))
            if idx < self.count() and idx >= 0:
                ingrid = True
            else:
                ingrid = False
        elif isinstance(idx, np.ndarray):
            ix = np.mod(np.mod(idx, self.nx * self.ny), self.nx).astype(int)
            iy = np.floor(np.mod(idx, self.nx * self.ny) / self.nx).astype(int)
            iz = np.floor(idx / (self.nx * self.ny)).astype(int)
            ingrid = np.logical_and(idx >= 0, idx < self.count())
        else:
            raise TypeError(('ix, iy and iz should be the same type, which is either'
                             'int or numpy.ndarray of matching length'))
        return ix, iy, iz, ingrid

    def get_index(self, x, y, z):
        """
        Will return the 0-indexed 1-Dimensional grid index given 3 coordinates as well
        as a True/False indicator for inside or outside the grid definition.

        Uses:
            pygeostat.get_index3d()

        Parameters:
            x (float or numpy.ndarray): x-coordinate value or numpy array of x-coordinate values
            y (float or numpy.ndarray): y-coordinate value or numpy array of y-coordinate values
            z (float or numpy.ndarray): z-coordinate value or numpy array of z-coordinate values

        Returns:
            idx (int or numpy.ndarray): 1-d index of the grid block containing the coordinates
            ingrid (bool or numpy.ndarray): in (True) or out (False) of the grid

        Examples:

            Calculate a 1d grid index based on an input coordinate (floats). Returns the index as an
            integer, as well as a boolean of whether the coordinate is in the grid:

            >>> x, y, z = 30.5, 12.5, 0.5
            >>> idx, ingrid = griddef.gridIndexCoords(x, y, z)

            Calculate 1d grid indices based on input coordinates (numpy arrays). Returns the indices
            as an array, as well as a boolean array of whether the coordinates are in the grid:

            >>> idx = np.array([0, 230, 460, 690, 920])
            >>> ix, iy, iz, ingrid = griddef.index1d_to_index3d(idx)
    """
        # Get the indices
        ix, iy, iz, ingrid = self.get_index3d(x, y, z)
        if isinstance(ix, np.ndarray):
            idx = ix + iy * self.nx + iz * self.nx * self.ny
            # Handle values outside the grid
            idx = np.where(ingrid, idx, -1)
            idx = idx.astype(int)
        else:
            if ingrid:
                idx = int(ix + iy * self.nx + iz * self.nx * self.ny)
            else:
                idx = -1
        return idx, ingrid

    def get_index3d(self, x, y, z):
        """Retruns the 0-indexed 3-Dimensional grid indices given 3 coordinates as well as a
        True/False indicator for inside or outside the grid definition.

        Parameters:
            x (float or numpy.ndarray): x-coordinate value or numpy array of x-coordinate values
            y (float or numpy.ndarray): y-coordinate value or numpy array of y-coordinate values
            z (float or numpy.ndarray): z-coordinate value or numpy array of z-coordinate values

        Returns:
            ix (int or numpy.ndarray): x index of the grid block
            iy (int or numpy.ndarray): y index of the grid block
            iz (int or numpy.ndarray): z index of the grid block
            ingrid (bool or numpy.ndarray): in (True) or out (False) of the grid

        Examples:

            Calculate a 3d grid index based on an input coordinate (floats). Returns the index as
            integers, as well as a boolean of whether the coordinate is in the grid:

            >>> x, y, z = 30.5, 12.5, 0.5
            >>> ix, iy, iz, ingrid = griddef.get_index3d(x, y, z)

            Calculate 3d grid indices based on input coordinates (numpy arrays). Returns the indices
            as arrays, as well as a boolean array of whether the coordinates are in the grid:

            >>> x, y = np.linspace(30.5, 100.5, 5), np.linspace(30.5, 100.5, 5)
            >>> ix, iy, iz, ingrid = griddef.get_index3d(x, y, z)

        """
        if hasattr(x, 'values'):
            x = x.values
        if hasattr(y, 'values'):
            y = y.values
        if hasattr(z, 'values'):
            z = z.values
        if all(isinstance(c, np.ndarray) for c in [x, y, z]):
            # Vectorized ix, iy, iz calcs
            ix = np.ceil((x - self.xmn) / self.xsiz + 0.5)
            iy = np.ceil((y - self.ymn) / self.ysiz + 0.5)
            iz = np.ceil((z - self.zmn) / self.zsiz + 0.5)
            # Check if on the minimum edge
            ix = np.where((ix == 0) & (x == self.xmn - (self.xsiz / 2.0)), 1, ix)
            iy = np.where((iy == 0) & (y == self.ymn - (self.ysiz / 2.0)), 1, iy)
            iz = np.where((iz == 0) & (z == self.zmn - (self.zsiz / 2.0)), 1, iz)
            # Check if within the grid
            ingrid = np.where((ix > 0) & (ix <= self.nx) & (iy > 0) & (
                iy <= self.ny) & (iz > 0) & (iz <= self.nz), True, False)
            # Get the final indices
            ix = ix - 1
            iy = iy - 1
            iz = iz - 1
            # Handle values outside the grid
            ix = np.where(ingrid, ix, -1)
            ix = ix.astype(int)
            iy = np.where(ingrid, iy, -1)
            iy = iy.astype(int)
            iz = np.where(ingrid, iz, -1)
            iz = iz.astype(int)
        else:  # default unvectorized version
            ix = np.ceil((x - self.xmn) / self.xsiz + 0.5)
            iy = np.ceil((y - self.ymn) / self.ysiz + 0.5)
            iz = np.ceil((z - self.zmn) / self.zsiz + 0.5)
            # Check if on minimum edge,
            if ix == 0 and x == self.xmn - (self.xsiz / 2.0):
                ix = 1
            if iy == 0 and y == self.ymn - (self.ysiz / 2.0):
                iy = 1
            if iz == 0 and z == self.zmn - (self.zsiz / 2.0):
                iz = 1
            # Check if within the grid
            if ix > 0 and ix <= self.nx and iy > 0 and iy <= self.ny and iz > 0 and iz <= self.nz:
                ix = int(ix - 1)
                iy = int(iy - 1)
                iz = int(iz - 1)
                ingrid = True
            else:
                ix = -1
                iy = -1
                iz = -1
                ingrid = False
        return ix, iy, iz, ingrid

    def get_coordinates(self, ix=None, iy=None, iz=None, idx=None):
        """Returns the 3 coordinate values based on the passed grid index. If no indices are
        passed, then the coordinates of all grid nodes are returned. Either all 3-D indices
        must all be passed (ix, iy, iz), the 1-D index is passed (idx), or all kwargs are None
        (returns all coordinates).

        Parameters:
            ix (int): x index
            iy (int): y index
            iz (int): z index
            idx (int) : 1-D index

        Returns:
            x (float or numpy.ndarray): x-coordinate value, or values if all grid nodes are returned
            y (float or numpy.ndarray): y-coordinate value, or values if all grid nodes are returned
            z (float or numpy.ndarray): z-coordinate value, or values if all grid nodes are returned

        Note:
            The option to return all grid node coordinates is  memory intensive for > 60 M cell grids.

        Usage:

            Generate a grid definition, and generate the coordinate arrays:

            >>> griddef = gs.GridDef(gridstr="50 0.5 1 \\n50 0.5 1 \\n50 0.5 1")
            >>> x, y, z = griddef.get_coordinates()

            Generate coordinates (floats) corresponding with a specific 3d index:

            >>> x, y, z = griddef.get_coordinates(1, 1, 1)

        """
        if all([test is not None for test in [ix, iy, iz]]):
            if idx is not None:
                raise ValueError('ix, iy and iz must be None if idx is not None (or vice versa)')
            x = ix * self.xsiz + self.xmn
            y = iy * self.ysiz + self.ymn
            z = iz * self.zsiz + self.zmn
        elif idx is not None:
            if any([test is not None for test in [ix, iy, iz]]):
                raise ValueError('ix, iy and iz must be None if idx is not None (or vice versa)')
            ix, iy, iz, _ = self.index1d_to_index3d(idx)
            x = ix * self.xsiz + self.xmn
            y = iy * self.ysiz + self.ymn
            z = iz * self.zsiz + self.zmn
        elif all([test is None for test in [ix, iy, iz, idx]]):
            xmn = self.xmn
            xmx = xmn + self.xsiz * (self.nx - 1)
            x = np.linspace(xmn, xmx, self.nx)
            ymn = self.ymn
            ymx = ymn + self.ysiz * (self.ny - 1)
            y = np.linspace(ymn, ymx, self.ny)
            zmn = self.zmn
            zmx = zmn + self.zsiz * (self.nz - 1)
            z = np.linspace(zmn, zmx, self.nz)
            # Things get a bit weird here. Python does y-fastest so the X and Y are switched here
            y, x, z = np.meshgrid(y, x, z, indexing='xy')
            nshp = self.count()
            x = np.reshape(x, nshp, order='F')
            y = np.reshape(y, nshp, order='F')
            z = np.reshape(z, nshp, order='F')
        else:
            raise ValueError('invalid kwarg combination - refer to the docstring!')
        return x, y, z

    def get_vertical_indices(self, x, y):
        """Returns grid indices corresponding with drilling a vertical
        drill hole intersecting all z blocks on the way down

        Parameters:
            x (float): x-coordinate value
            y (float): y-coordinate value

        Returns:
            indices (dict): grid indices of vertical 'drill hole'"""
        # Get the extents
        [(xmin, xmax), (ymin, ymax), (zmin, zmax)] = self.extents()
        # Check if we are within the bounds of the grid
        if x < xmin or x > xmax or y < ymin or y > ymax:
            return []
        z = np.linspace(zmin + self.zsiz * 0.5, zmax - self.zsiz * 0.5, num=self.nz)
        x, y = x * np.ones(z.shape), y * np.ones(z.shape),
        idx, _ = self.get_index(x, y, z)
        return idx

    def get_slice_coordinate(self, orient, index):
        """Returns the real coordinate for a slice given the index in the grid and orientation
        NOTE: assumes 0-indexed slice coordinates are passed.

        Parameters:
            orient (str) : orientation of the current slice ('xy', 'yz', 'xz')
            index (int) : index of the current slice

        Returns:
            cord (float) : coordinate of the current slice

        Usage:
        >>> griddef.get_slice_coordinate('xy',10)

        """
        if orient == 'xy':  # want the z coordinate
            cord = self.zmn + self.zsiz * index
        elif orient == 'yz':  # want the x coordinate
            cord = self.xmn + self.xsiz * index
        elif orient == 'xz':  # want the y coordinate
            cord = self.ymn + self.ysiz * index

        return cord

    def get_slice_index(self, orient, coordinate):
        """Returns the index in the grid along the correct dimension for the specificed slice
        i.e. the z index for an 'xy' slice

        >>> griddef.get_slice_index('xy',700.2)

        Parameters:
            orient (str) : orientation of the current slice ('xy', 'yz', 'xz')
            coordinate (float) : index of the current slice

        Returns:
            index (int) : index to the specified slice

        """
        if orient == 'xy':  # want the z coordinate
            ix, iy, iz, infl = self.get_index3d(self.xmn, self.ymn, coordinate)
            index = iz
        elif orient == 'yz':  # want the x coordinate
            ix, iy, iz, infl = self.get_index3d(coordinate, self.ymn, self.zmn)
            index = ix
        elif orient == 'xz':  # want the y coordinate
            ix, iy, iz, infl = self.get_index3d(self.xmn, coordinate, self.zmn)
            index = iy
        return index

    def change_blocknum(self, nx_new, ny_new, nz_new, return_copy=False):
        """
        Function to change the blocksize of the current grid while retaining the original bounding
        box.

        Useful if attempting to work at a coarse grid (for speed) prior to obtaining a final
        estimate at the original resolution.

        Parameters:
            nx_new (int): New number of blocks in X direction
            ny_new (int): New number of blocks in Y direction
            nz_new (int): New number of blocks in Z direction

        Example:
            Define a grid:

            >>> import pygeostat as gs
            >>> grid = gs.GridDef(grid_str="100 5 10  \\n100 5 10  \\n100 5 5")

            Change the dimensions of the grid:

            >>> grid.changedim(50,50,50)
            >>> print(grid.nx,grid.xmn,grid.xsiz)
            >>> print(grid.ny,grid.ymn,grid.ysiz)
            >>> print(grid.nz,grid.zmn,grid.zsiz)

            Use the changed resolution grid in a parameter file:

            >>> parstr = "TestParFile \\n{grd}"
            >>> prog = gs.Program(program='./text.exe',parstr=parstr.format(grd=str(grid)))

        """
        bounds = np.array(self.extents())
        span = np.array([bounds[0, 1] - bounds[0, 0],
                         bounds[1, 1] - bounds[1, 0],
                         bounds[2, 1] - bounds[2, 0]])
        newsize = span / np.array([nx_new, ny_new, nz_new])
        newmin = bounds[:, 0] + 0.5 * newsize
        # assign the new dimensions to the class
        nx = int(nx_new)
        ny = int(ny_new)
        nz = int(nz_new)
        xmn = newmin[0]
        ymn = newmin[1]
        zmn = newmin[2]
        xsiz = newsize[0]
        ysiz = newsize[1]
        zsiz = newsize[2]

        if return_copy is False:
            self.__init__(grid_arr=[nx, xmn, xsiz, ny, ymn, ysiz, nz, zmn, zsiz])
        else:
            from . import GridDef as GridDef
            return GridDef(grid_arr=[nx, xmn, xsiz, ny, ymn, ysiz, nz, zmn, zsiz])

    def change_blocksize(self, xsiz_new, ysiz_new, zsiz_new, return_copy=False):
        """
        Function to change the size of individual blocks in the grid.
        Finds the new number of blocks
        given the target sizes in each direction.

        Parameters:
            xsiz_new (float): New size of blocks in X
            ysiz_new (float): New size of blocks in Y
            zsiz_new (float): New size of blocks in Z
            return_copy (bool): if True will return a copy instead of modifying self

        """
        bounds = np.array(self.extents())
        span = np.array([bounds[0, 1] - bounds[0, 0],
                         bounds[1, 1] - bounds[1, 0],
                         bounds[2, 1] - bounds[2, 0]])
        newblocks = np.ceil(span / np.array([xsiz_new, ysiz_new, zsiz_new]))
        newmin = bounds[:, 0] + 0.5 * np.array([xsiz_new, ysiz_new, zsiz_new])
        # assign the new dimensions to the class
        nx = int(newblocks[0])
        ny = int(newblocks[1])
        nz = int(newblocks[2])
        xmn = newmin[0]
        ymn = newmin[1]
        zmn = newmin[2]
        xsiz = xsiz_new
        ysiz = ysiz_new
        zsiz = zsiz_new

        if return_copy is False:
            self.__init__(grid_arr=[nx, xmn, xsiz, ny, ymn, ysiz, nz, zmn, zsiz])
        else:
            from . import GridDef as GridDef
            return GridDef(grid_arr=[nx, xmn, xsiz, ny, ymn, ysiz, nz, zmn, zsiz])

    def generate_grid_points(self):
        """
        Returns a (ncell x 3) set of real location grid points that conform to the standard gslib
        orderings, starting at the bottom of the model and increasing upwards, x-fastest, then y,
        then z.

        Note:
            Memory intensive for > 60 M cell grids. This function likely rarely needs to be used,
            and does not explicitly account for a mask in construction. Can be used to get the set
            of GSLIB evaluation locations from a griddef, for example.

        Usage:
            Generate a grid definition, and generate the array of locations:

            >>> griddef = gs.GridDef(grid_str="50 0.5 1 \\n50 0.5 1 \\n50 0.5 1")
            >>> gridxyz = griddef.generate_grid_points()

        """
        xmin = self.xmn
        xmax = xmin + self.xsiz * (self.nx - 1)
        xrng = np.linspace(xmin, xmax, self.nx)

        ymin = self.ymn
        ymax = ymin + self.ysiz * (self.ny - 1)
        yrng = np.linspace(ymin, ymax, self.ny)

        zmin = self.zmn
        zmax = zmin + self.zsiz * (self.nz - 1)
        zrng = np.linspace(zmin, zmax, self.nz)

        # Python does y-fastest so the X and Y are switched here
        y1, x1, z1 = np.meshgrid(yrng, xrng, zrng, indexing='xy')
        nshp = self.count()
        x = np.reshape(x1, nshp, order='F')
        y = np.reshape(y1, nshp, order='F')
        z = np.reshape(z1, nshp, order='F')
        xyz = np.c_[x, y, z]
        return xyz

    def pad_grid(self, nx_pad, ny_pad, nz_pad, return_copy=False):
        """
        Pad the grid on either side in all directions by the number of cells specified on input

        Parameters:
            nx_pad (int or tuple): number of cells in x direction to add to the grid on each side
            ny_pad (int or tuple): number of cells in y direction to add to the grid on each side
            nz_pad (int or tuple): number of cells in z direction to add to the grid on each side
            return_copy (bool): return copy or reinitialize self

        Examples:

            Generate a grid definition:

            >>> griddef = gs.GridDef(gridstr="50 0.5 1 \\n50 0.5 1 \\n1 0.5 1")

            Symmetrically pad the grid cells in the x and y directions

            >>> griddef2 = griddef2.padgrid(10, 10, 0, return_copy=True)

            Asymmetrically pad the grid with cells in the x and y directions

            >>> griddef2.padgrid((6, -5), 5, 0)

        """
        if isinstance(nx_pad, int):
            xmn = self.xmn - nx_pad * self.xsiz
            nx = self.nx + 2 * nx_pad
        elif isinstance(nx_pad, tuple):
            xmn = self.xmn - int(nx_pad[0]) * self.xsiz
            nx = self.nx + int(nx_pad[0] + nx_pad[1])
        else:
            raise ValueError('nx_pad must be either an integer or tuple')

        if isinstance(ny_pad, int):
            ymn = self.ymn - ny_pad * self.ysiz
            ny = self.ny + 2 * ny_pad
        elif isinstance(nx_pad, tuple):
            ymn = self.ymn - int(ny_pad[0]) * self.ysiz
            ny = self.ny + int(ny_pad[0] + ny_pad[1])
        else:
            raise ValueError('nx_pad must be either an integer or tuple')

        if isinstance(nz_pad, int):
            zmn = self.zmn - nz_pad * self.zsiz
            nz = self.nz + 2 * nz_pad
        elif isinstance(nx_pad, tuple):
            zmn = self.zmn - int(nz_pad[0]) * self.zsiz
            nz = self.nz + int(nz_pad[0] + nz_pad[1])
        else:
            raise ValueError('nx_pad must be either an integer or tuple')

        # reinitialize this object with the new parameters

        if return_copy is False:
            self.__init__(grid_arr=[nx, xmn, self.xsiz, ny, ymn, self.ysiz, nz, zmn, self.zsiz])
        else:
            from . import GridDef as GridDef
            return GridDef(grid_arr=[nx, xmn, self.xsiz, ny, ymn, self.ysiz, nz, zmn, self.zsiz])

    def random_points(self, n=100):
        """
        Generate random points from within the grid
        """
        x = self.xmn + np.random.rand(n) * self.nx * self.xsiz
        y = self.ymn + np.random.rand(n) * self.ny * self.ysiz
        if self.nz > 1:
            z = self.zmn + np.random.rand(n) * self.nz * self.zsiz
            return np.c_[x, y, z]
        else:
            return np.c_[x, y]

    def random_indices(self, dim=3, n=20, seed=None, buffer_x=None, buffer_y=None, buffer_z=None):
        """
        Generate a list of random indices from within the grid

        Parameters:
            dim (int): 1 returns a 1D index, 3 returns x, y, and z indexes.
            n (int): The number of indices to return for each dimension
            seed (int): If provided will initialized the random number generator with the seed.
                If not provided then you will get different values every time you run this function
            buffer_x (int): you can set a buffer if you don't want any random indices near
                the edge of the x border of the grid, Note: only works for dim=3
            buffer_y (int): you can set a buffer if you don't want any random indices near
                the edge of the y border of the grid, Note: only works for dim=3
            buffer_z (int): you can set a buffer if you don't want any random indices near
                the edge of the z border of the grid, Note: only works for dim=3

        Returns:
            "dim = 1"
            indice (list): a 1D indice of size `n`
            "dim = 3"
            xind (list): a list of indices in x dimension of size `n`
            yind (list): a list of indices in y dimension of size `n`
            zind (list): a list of indices in z dimension of size `n`
        """
        from .. utility.logging import printerr

        if seed:
            if isinstance(seed, int):
                np.random.seed(seed)
            else:
                raise ValueError('seed must be an integer')

        if buffer_x:
            if isinstance(buffer_x, int):
                bx = (buffer_x, buffer_x)
            elif isinstance(buffer_x, tuple):
                if isinstance(buffer_x[0], int) and isinstance(buffer_x[1], int):
                    bx = buffer_x
            else:
                raise ValueError('buffer_x is not an integer or a tuple of integers', buffer_x)

        if buffer_y:
            if isinstance(buffer_y, int):
                by = (buffer_y, buffer_y)
            elif isinstance(buffer_y, tuple):
                if isinstance(buffer_y[0], int) and isinstance(buffer_y[1], int):
                    by = buffer_y
            else:
                raise ValueError('buffer_y is not an integer or a tuple of integers', buffer_y)

        if buffer_z:
            if isinstance(buffer_z, int):
                bz = (buffer_z, buffer_z)
            elif isinstance(buffer_z, tuple):
                if isinstance(buffer_z[0], int) and isinstance(buffer_z[1], int):
                    bz = buffer_z
            else:
                raise ValueError('buffer_z is not an integer or a tuple of integers', buffer_z)

        if dim != 1:
            if dim != 3:
                printerr('Only 1 or 3 Dimensions allowed. Assuming you wanted 3 dimension',
                         errtype='warning')
                print('dim entered = ', dim)
            if buffer_x:
                xrange = (0 + bx[0], self.nx - bx[1])
                if xrange[0] > xrange[1] or xrange[0] == xrange[1]:
                    raise ValueError('buffer_x is to large for the grid causing the upper range'
                                     ' to be less then or equal to the lower range', xrange)
            else:
                xrange = (0, self.nx)
            if buffer_y:
                yrange = (0 + by[0], self.ny - by[1])
                if yrange[0] > yrange[1] or yrange[0] == yrange[1]:
                    raise ValueError('buffer_y is to large for the grid causing the upper range'
                                     '  to be less then or equal to the lower range', yrange)
            else:
                yrange = (0, self.ny)
            if buffer_z:
                zrange = (0 + bz[0], self.nz - bz[1])
                if zrange[0] > zrange[1] or zrange[0] == zrange[1]:
                    raise ValueError('buffer_z is to large for the grid causing the upper range'
                                     ' to be less then or equal to the lower range', zrange)
            else:
                zrange = (0, self.nz)
            rand_xind = np.random.random_integers(xrange[0], xrange[1], size=n)
            rand_yind = np.random.random_integers(yrange[0], yrange[1], size=n)
            rand_zind = np.random.random_integers(zrange[0], zrange[1], size=n)
        else:
            rand_indices = np.random.random_integers(0, self.count(), size=n)

        if dim == 1:
            return rand_indices
        else:
            return rand_xind, rand_yind, rand_zind



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Various tools for calculating statistics
"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
import numpy as np


def weighted_mean(var, wts):
    """Calculates the weighted mean"""
    return np.average(var, weights=wts)


def weighted_variance(var, wts):
    """Calculates the weighted variance"""
    return np.average((var - weighted_mean(var, wts))**2, weights=wts)


def weighted_skew(var, wts):
    """Calculates the weighted skewness"""
    return (np.average((var - weighted_mean(var, wts))**3, weights=wts) /
            weighted_variance(var, wts)**(1.5))

def weighted_kurtosis(var, wts):
    """Calculates the weighted skewness"""
    return (np.average((var - weighted_mean(var, wts))**4, weights=wts) /
            weighted_variance(var, wts)**(2))

def weighted_covariance(x, y, wt):
    """Calculates the weighted covariance"""
    return (np.average((x - weighted_mean(x, wt)) *
                       (y - weighted_mean(y, wt)), weights=wt))


def weighted_correlation(x, y, wt):
    """Calculates the weighted correlation"""
    return (weighted_covariance(x, y, wt) /
            (np.sqrt(weighted_variance(x, wt)) * np.sqrt(weighted_variance(y, wt))))


def weighted_correlation_rank(x, y, wt):
    """Calculatees the weighted spearman rank correlation"""
    from scipy.stats import rankdata
    x = rankdata(x)
    y = rankdata(y)
    return weighted_correlation(x, y, wt)

def near_positive_definite(input_matrix):
    """This function uses R to calculate the nearest positive definite matrix
    within python. An installation of R with the library "Matrix" is required.
    The module rpy2 is also needed

    The only requirement is an input matrix. Can be either a pandas dataframe or
    numpy-array.

    Parameters:
        input_matrix:  input numpy array or pandas dataframe, not numpy matrix

    Returns:
        (np.array):  Nearest positive definite matrix as a numpy-array
    """
    import pandas as pd
    import numpy as np
    from ..utility.logging import printerr
    import pygeostat as gs
    # Try and load r2py
    try:
        import rpy2.robjects as robjects
        from rpy2.robjects import r
        from rpy2.robjects.packages import importr
    except ImportError:
        printerr(("near_positive_definite could not be loaded. Please install the r2py library"
                     " and the software R with the library 'Matrix' to enable it. Installation"
                     " instructions can be found within pygeostat's documentation."),
                    errtype='error')
        return
    # Convert input matrix to a numpy array if it is a pd.DataFrame
    if isinstance(input_matrix, pd.DataFrame):
        input_matrix = input_matrix.as_matrix()
    # Determine matrix shape
    dim = input_matrix.shape
    # Call matrix R library
    matcalc = importr("Matrix")
    # Convert numpy array to RObject then to R matrix
    pdmat = robjects.FloatVector(input_matrix.reshape((input_matrix.size)))
    pdmat = robjects.r.matrix(pdmat, nrow=dim[0], ncol=dim[1], byrow=True)
    # Calculate nearest positive definite matrix
    pdmat = matcalc.near_positive_definite(pdmat)
    # Convert calculated matrix to python string
    pdmat = pdmat[0]  # Extract near_positive_definite matrix from R object
    pdmat = r.toString(pdmat)  # Convert R binary object to a string
    pdmat = pdmat.r_repr()  # Convert R string to python string
    pdmat = pdmat.replace('"', "")  # Clean up string
    pdmat = pdmat.replace(' ', "")  # Clean up string
    # Convert comma delimited string to list then to np array
    pdmat = [float(x) for x in pdmat.split(',')]
    pdmat = np.array(pdmat)
    pdmat = np.reshape(pdmat, dim)  # Restore near_positive_definite matrix to the original input shape

    return pdmat


def accsim(truth, reals, pinc=0.05):
    """
    Calculates the proportion of locations where the true value falls within symmetric p-PI
    intervals when completing a jackknife study. A portion of the data is excluded from the
    conditioning dataset and the excluded sample locations simulated values are then checked.

    .. seealso::

        Pyrcz, M. J., & Deutsch, C. V. (2014). Geostatistical Reservoir Modeling (2nd ed.). New
        York, NY: Oxford University Press, p. 350-351.

    Arguments:
        truth: Tidy (long-form) 1D data where a single column containing the true values.
            A pandas dataframe/series or numpy array can be passed
        reals: Tidy (long-form) 2D data where a single column contains values from a single
            realizations and each row contains the simulated values from a single truth location.
            A pandas dataframe or numpy matrix can be passed

    Keyword Arguments:
        pinc (float): Increments between the probability intervals to calculate within (0, 1)

    Returns:
        propavg (pd.DataFrame): Dataframe with the calculated probability intervals and the
        fraction within the interval

    Returns:
        sumstats (dict): Dictionary containing the average variance (U), mean squared error (MSE),
        accuracy measure (acc), precision measure (pre), and a goodness measure (goo)

    """
    import pandas as pd
    import pygeostat as gs
    # Handle input
    if isinstance(truth, pd.Series):
        truth = truth.values
    elif isinstance(truth, pd.DataFrame):
        truth = truth.values
    elif not isinstance(truth, np.ndarray):
        raise ValueError("The argument `truth` must be a pd.DataFrame, pd.Series, or np.matrix")
    if isinstance(truth, np.ndarray) and len(truth.shape) == 1:
        truth = np.reshape(truth, (truth.shape[0], 1))
    if isinstance(reals, pd.DataFrame):
        reals = reals.values
    elif not isinstance(reals, np.ndarray):
        raise ValueError("The argument `reals` must be a pd.DataFrame or np.matrix")
    try:
        data = np.concatenate((truth, reals), axis=1)
        data = pd.DataFrame(data=data)
    except:
        raise ValueError("The `truth` and `reals` data could not be coerced into a pd.DataFrame")
    # Initialize some variables
    pints = np.arange(pinc, 1, pinc)
    propindic = dict([pint, []] for pint in pints)
    variances = []
    acc = dict([pint, 0] for pint in pints)
    pre = dict([pint, 0] for pint in pints)
    goo = dict([pint, 0] for pint in pints)
    # Calculate the indicator responses and local variances
    for i, values in data.iterrows():
        cdf = gs.cdf(values[1:].values)
        variances.append(np.var(values[1:].values))
        for pint in pints:
            if cdf[0][0] <= values[0] <= cdf[0][-1]:
                p = gs.z_percentile(values[0], cdf[0], cdf[1])
                plower = 0.5 - (pint / 2)
                pupper = 0.5 + (pint / 2)
                if plower <= p <= pupper:
                    indic = 1
                else:
                    indic = 0
            else:
                indic = 0
            propindic[pint].append(indic)
    # Calculate the average proportions and average variance
    propavg = []
    for pint in pints:
        avg = np.average(propindic[pint])
        propavg.append([pint, avg])
    propavg = pd.DataFrame(propavg, columns=['ProbInt', 'FracIn'])
    # Calculate the summary statistics
    avgvar = np.average(variances)
    mse = ((propavg['ProbInt'].values - propavg['FracIn'].values) ** 2).mean()
    acc = 0
    pre = 0
    goo = 0
    for i, values in propavg.iterrows():
        if values[1] >= values[0]:
            acc = acc + 1
            pre = pre + (values[1] - values[0])
            goo = goo + (values[1] - values[0])
        else:
            goo = goo + (2 * (values[0] - values[1]))
    acc = acc / len(propavg)
    pre = 1 - ((2 * pre) / len(propavg))
    goo = 1 - (goo / len(propavg))
    sumstats = {'avgvar': avgvar, 'mse': mse, 'acc': acc, 'pre': pre, 'goo': goo}

    return propavg, sumstats

def accmik(truth, thresholds, mikprobs, pinc=0.05):
    """
    Similar to accsim but accepting mik distributions instead

    Mostly pulled from accsim

    Parameters
    ----------
        truth: np.ndarray
            Tidy (long-form) 1D data where a single column containing the true values.
            A pandas dataframe/series or numpy array can be passed
        thresholds: np.ndarray
            1D array of thresholds where each CDF is defined by these thresholds and the probability
            given in the mikprobs array for each location.
        mikprobs: np.ndarray
            Tidy (long-form) 2D data where a single column contains values from a single
            MIK cutoff and each row contains the simulated values for the corresponding single
            truth location. A pandas dataframe or numpy matrix can be passed
        pinc: float
            Increments between the probability intervals to calculate within (0, 1)

    Returns
    -------
        propavg: pd.DataFrame
            Dataframe with the calculated probability intervals and the fraction within the interval
        sumstats: dict
            Dictionary containing the average variance (U), mean squared error (MSE), accuracy
            measure (acc), precision measure (pre), and a goodness measure (goo)

    """
    import pandas as pd
    # Handle input
    if isinstance(truth, pd.Series):
        truth = truth.values
    elif isinstance(truth, pd.DataFrame):
        truth = truth.values
    elif not isinstance(truth, np.ndarray):
        raise ValueError("The argument `truth` must be a pd.DataFrame, pd.Series, or np.matrix")
    if isinstance(truth, np.ndarray) and len(truth.shape) == 1:
        truth = np.reshape(truth, (truth.shape[0], 1))
    if isinstance(mikprobs, pd.DataFrame):
        mikprobs = mikprobs.values
    elif not isinstance(mikprobs, np.ndarray):
        raise ValueError("The argument `mikprobs` must be a pd.DataFrame or np.matrix")
    # Initialize some variables
    pints, propindic, variances = _interval_responses(truth, mikprobs, pinc, cdf_x=thresholds)
    # Calculate the average proportions and average variance
    propavg = []
    for pint in pints:
        avg = np.average(propindic[pint])
        propavg.append([pint, avg])
    propavg = pd.DataFrame(propavg, columns=['ProbInt', 'FracIn'])
    # Calculate the summary statistics
    avgvar = np.average(variances)
    mse = ((propavg['ProbInt'].values - propavg['FracIn'].values) ** 2).mean()
    acc = 0
    pre = 0
    goo = 0
    for i, values in propavg.iterrows():
        if values[1] >= values[0]:
            acc = acc + 1
            pre = pre + (values[1] - values[0])
            goo = goo + (values[1] - values[0])
        else:
            goo = goo + (2 * (values[0] - values[1]))
    acc = acc / len(propavg)
    pre = 1 - ((2 * pre) / len(propavg))
    goo = 1 - (goo / len(propavg))
    sumstats = {'avgvar': avgvar, 'mse': mse, 'acc': acc, 'pre': pre, 'goo': goo}
    return propavg, sumstats


def _interval_responses(truth, reals, pinc, cdf_x=None):
    """
    When cdf_x is None, reals contains the simulated values from which a cdf should be computed.
    Otherwise the ``'reals'`` contains the distribution F(cdf_x) values for each location (nloc,
    nquant)

    Mostly pulled from the original accsim

    Parameters:

        truth: np.ndarray
            tidy 1D array of truth values
        reals: np.ndarray
            tidy 2D array of `reals` where if cdf_x is None these are the realizations from which a cdf
            is built, otherwise cdf_x defines the z-values and each row of reals contains the
            corresponding probabilites defining the local cdf's
        pinc: float
            the incremement of the probability intervals
        cdf_x: np.ndarray, optional
            contains the z-values when ``reals`` contains the F(z)

    Returns:
        pints: np.ndarray
            a range of pinc spaced probability intervals
        propindic: dict
            the dictionary used in accsim and accmik functions
        variances: list
            the list of variances
    """
    from .cdf import variance_from_cdf
    if not isinstance(truth, np.ndarray):
        truth = np.array(truth)
    isjagged = False
    if isinstance(reals, list) or reals.dtype == "O":
        # assume reals is `jagged` (nested lists), each location has a different # simulated vals
        isjagged = True
    elif not isinstance(reals, np.ndarray):
        reals = np.array(reals)
    if truth.shape[0] != reals.shape[0]:
        raise ValueError('`truth` and `reals` must have the same dimension along the first axis!')
    # initializse the variables
    pints = np.arange(pinc, 1, pinc)
    propindic = {pint: [] for pint in pints}
    variances = []
    if cdf_x is not None and reals[0, 0] != 0:
        reals = np.c_[np.ones(truth.shape[0]), reals]
        cdf_x = np.insert(cdf_x, [0], cdf_x[0] - (cdf_x[1] - cdf_x[0]))
    # Calculate the indicator responses and local variances
    for i in range(truth.shape[0]):
        if cdf_x is None:
            if isjagged:
                ecdf = cdf(reals[i])  # each element in reals is a list of sim vals
                v = np.var(reals[i])
            else:
                ecdf = cdf(reals[i, :])  # reals is a 2D array with standard size that can be sliced
                v = np.var(reals[i, :])
            variances.append(v)
        else:
            ecdf = (cdf_x, reals[i, :])
            variances.append(variance_from_cdf(ecdf[0], ecdf[1]))
        for pint in pints:
            if ecdf[0][0] <= truth[i] <= ecdf[0][-1]:
                p = z_percentile(truth[i], ecdf[0], ecdf[1])
                plower = 0.5 - (pint / 2)
                pupper = 0.5 + (pint / 2)
                if plower <= p <= pupper:
                    indic = 1
                else:
                    indic = 0
            else:
                indic = 0
            propindic[pint].append(indic)
    return pints, propindic, variances



"""
The :mod:`pygslib.statistics` module includes statistics functions.
"""
from __future__ import absolute_import, division, print_function

from .kde import *
from .cdf import *
from .utils import *
from .postsim  import postsim_multfiles



#
# ATTRIBUTATION TO MATPLOTLIB:
# The template for the PygeostatDeprecationWarning class and related functions within this
# deprecation module is adapted from the MatplotLibDeprecationWarning and related module,
# as it provides a flexible and extensible approach to deprecation warnings.
#
# The Matplotlib License is is available at: https://matplotlib.org/users/license.html,
# which was accessed on March 26, 2018. Relevant to this application are the items:
#
# 1. This LICENSE AGREEMENT is between the Matplotlib Development Team (“MDT”), and the
# Individual or Organization (“Licensee”) accessing and otherwise using matplotlib software
# in source or binary form and its associated documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, MDT hereby grants Licensee
# a nonexclusive, royalty-free, world-wide license to reproduce, analyze, test, perform and/or
# display publicly, prepare derivative works, distribute, and otherwise use matplotlib 2.2.2 alone
# or in any derivative version, provided, however, that MDT’s License Agreement and MDT’s notice
# of copyright, i.e., “Copyright (c) 2012-2013 Matplotlib Development Team; All Rights Reserved”
# are retained in matplotlib 2.2.2 alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates
# matplotlib 2.2.2 or any part thereof, and wants to make the derivative work available
# to others as provided herein, then Licensee hereby agrees to include in any such work a
# brief summary of the changes made to matplotlib 2.2.2.
#
# Following this instruction, it should be noted that the name occurences of matplotlib
# were changed to pygeostat within this module. This was peformed to avoid confusion
# with the application of this deprecation warning within pygeostat, so it is not
# confused with deprecation relating to matplotlib. The underlying code, however,
# is entirely attributed to the Matplotlib Development team and is so accredited here.

import warnings
import functools


class PygeostatDeprecationWarning(UserWarning):
    """
    A class for issuing deprecation warnings for Pygeostat users.

    In light of the fact that Python builtin DeprecationWarnings are ignored
    by default as of Python 2.7 (see link below), this class was put in to
    allow for the signaling of deprecation, but via UserWarnings which are not
    ignored by default.

    https://docs.python.org/dev/whatsnew/2.7.html#the-future-for-python-2-x

    Attributation to Matplotlib:
    The template for the PygeostatDeprecationWarning class and related functions within this
    deprecation module is adapted from the MatplotLibDeprecationWarning and related module,
    as it provides a flexible and extensible approach to deprecation warnings. Refer to
    the header of deprecation.py for a full attributation and licence of Matplotlib.
    """
    pass


gsDeprecation = PygeostatDeprecationWarning


def _generate_deprecation_message(since, message='', name='',
                                  alternative='', pending=False,
                                  obj_type='attribute',
                                  addendum=''):

    if not message:

        if pending:
            message = (
                'The %(name)s %(obj_type)s will be deprecated in a '
                'future version.')
        else:
            message = (
                'The %(name)s %(obj_type)s was deprecated in version '
                '%(since)s.')

    altmessage = ''
    if alternative:
        altmessage = ' Use %s instead.' % alternative

    message = ((message % {
        'func': name,
        'name': name,
        'alternative': alternative,
        'obj_type': obj_type,
        'since': since}) +
        altmessage)

    if addendum:
        message += addendum

    return message


def warn_deprecated(
        since, message='', name='', alternative='', pending=False,
        obj_type='attribute', addendum=''):
    """
    Used to display deprecation warning in a standard way.

    Parameters
    ----------
    since : str
        The release at which this API became deprecated.

    message : str, optional
        Override the default deprecation message.  The format
        specifier `%(name)s` may be used for the name of the function,
        and `%(alternative)s` may be used in the deprecation message
        to insert the name of an alternative to the deprecated
        function.  `%(obj_type)s` may be used to insert a friendly name
        for the type of object being deprecated.

    name : str, optional
        The name of the deprecated object.

    alternative : str, optional
        An alternative function that the user may use in place of the
        deprecated function.  The deprecation warning will tell the user
        about this alternative if provided.

    pending : bool, optional
        If True, uses a PendingDeprecationWarning instead of a
        DeprecationWarning.

    obj_type : str, optional
        The object type being deprecated.

    addendum : str, optional
        Additional text appended directly to the final message.

    Examples
    --------

        Basic example::

            # To warn of the deprecation of "Pygeostat.name_of_module"
            warn_deprecated('1.4.0', name='Pygeostat.name_of_module',
                            obj_type='module')

    """
    message = _generate_deprecation_message(
        since, message, name, alternative, pending, obj_type)

    warnings.warn(message, gsDeprecation, stacklevel=1)


def deprecated(since, message='', name='', alternative='', pending=False,
               obj_type=None, addendum=''):
    """
    Decorator to mark a function or a class as deprecated.

    Parameters
    ----------
    since : str
        The release at which this API became deprecated.  This is
        required.

    message : str, optional
        Override the default deprecation message.  The format
        specifier `%(name)s` may be used for the name of the object,
        and `%(alternative)s` may be used in the deprecation message
        to insert the name of an alternative to the deprecated
        object.  `%(obj_type)s` may be used to insert a friendly name
        for the type of object being deprecated.

    name : str, optional
        The name of the deprecated object; if not provided the name
        is automatically determined from the passed in object,
        though this is useful in the case of renamed functions, where
        the new function is just assigned to the name of the
        deprecated function.  For example::

            def new_function():
                ...
            oldFunction = new_function

    alternative : str, optional
        An alternative object that the user may use in place of the
        deprecated object.  The deprecation warning will tell the user
        about this alternative if provided.

    pending : bool, optional
        If True, uses a PendingDeprecationWarning instead of a
        DeprecationWarning.

    addendum : str, optional
        Additional text appended directly to the final message.

    Examples
    --------

        Basic example::

            @deprecated('1.4.0')
            def the_function_to_deprecate():
                pass

    """

    def deprecate(obj, message=message, name=name, alternative=alternative,
                  pending=pending, addendum=addendum):
        import textwrap

        if not name:
            name = obj.__name__

        if isinstance(obj, type):
            obj_type = "class"
            old_doc = obj.__doc__
            func = obj.__init__

            def finalize(wrapper, new_doc):
                try:
                    obj.__doc__ = new_doc
                except (AttributeError, TypeError):
                    # cls.__doc__ is not writeable on Py2.
                    # TypeError occurs on PyPy
                    pass
                obj.__init__ = wrapper
                return obj
        else:
            obj_type = "function"
            if isinstance(obj, classmethod):
                func = obj.__func__
                old_doc = func.__doc__

                def finalize(wrapper, new_doc):
                    wrapper = functools.wraps(func)(wrapper)
                    wrapper.__doc__ = new_doc
                    return classmethod(wrapper)
            else:
                func = obj
                old_doc = func.__doc__

                def finalize(wrapper, new_doc):
                    wrapper = functools.wraps(func)(wrapper)
                    wrapper.__doc__ = new_doc
                    return wrapper

        message = _generate_deprecation_message(
            since, message, name, alternative, pending,
            obj_type, addendum)

        def wrapper(*args, **kwargs):
            warnings.warn(message, gsDeprecation, stacklevel=2)
            return func(*args, **kwargs)

        old_doc = textwrap.dedent(old_doc or '').strip('\n')
        message = message.strip()
        new_doc = (('\n.. deprecated:: %(since)s'
                    '\n    %(message)s\n\n' %
                    {'since': since, 'message': message}) + old_doc)
        if not old_doc:
            # This is to prevent a spurious 'unexected unindent' warning from
            # docutils when the original docstring was blank.
            new_doc += r'\ '

        return finalize(wrapper, new_doc)

    return deprecate



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''Contains utilities for GSLIB rotations'''
from __future__ import absolute_import, division, print_function

__author__ = 'pygeostat development team'
__date__ = '2016-02-14'
__version__ = '1.000'

import numpy as np
import pandas as pd
from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d import proj3d
import matplotlib.pyplot as plt

from ..data import DataFile

origin, xaxis, yaxis, zaxis = [0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]

def get_rotation_matrix(azm, dip, tilt):
    '''Returns the GSLIB rotation matrix given
       GSLIB angles for azimuth, dip and tilt'''
    from . import transformations as tf
    # Azimuth rotation about Z axis
    gamma = np.radians(azm - 90.0)
    Rz = tf.rotation_matrix(gamma, zaxis)
    # Dip rotation about Y' axis
    beta = np.radians(dip)
    Ry = tf.rotation_matrix(beta, yaxis)
    # Tilt rotation about X' axis
    alpha = np.radians(-1.0*tilt)
    Rx = tf.rotation_matrix(alpha, xaxis)
    # Combination of matrices
    R = tf.concatenate_matrices(Rx, Ry, Rz)
    # Skip the translation piece, just return the rotation matrix
    return(R[0:3,0:3])

def principalvectors(azm, dip, tilt, lefthandrule=True,
                     majorlen=1.0, minorlen=1.0, vertlen=1.0):
    '''Returns the GSLIB principal vectors:
    NOTE: we typically think of things/visualize using the left hand rule in
    which case the original axes are:
    major_axis, minor_axis, vert_axis = [1, 0, 0], [0, -1, 0], [0, 0, 1]
    although the GSLIB rotation matrix actually defines them as:
    major_axis, minor_axis, vert_axis = [1, 0, 0], [0, 1, 0], [0, 0, 1]
    This is only an issue in practice for an asymmetric minor axis'''
    # Rotation matrix - uses the transpose of the GSLIB rotation matrix
    rotmat = get_rotation_matrix(azm, dip, tilt).transpose()
    # Defined axes
    if lefthandrule:
        major_axis, minor_axis, vert_axis = [1, 0, 0], [0, -1, 0], [0, 0, 1]
    else:
        major_axis, minor_axis, vert_axis = [1, 0, 0], [0, 1, 0], [0, 0, 1]

    # Major - oriented + along y for azm=dip=tilt=0
    major_head = np.dot(rotmat, major_axis) * majorlen
    major_xs, major_ys, major_zs = (0, major_head[0]), (0, major_head[1]), (0, major_head[2])

    # Minor - oriented + along y for azm=dip=tilt=0 with Left Hand Rule
    #       - oriented - along y for azm=dip=tilt=0 with Right Hand Rule
    minor_head = np.dot(rotmat, minor_axis) * minorlen
    minor_xs, minor_ys, minor_zs = (0, minor_head[0]), (0, minor_head[1]), (0, minor_head[2])

    # Vert - oriented + along z for azm=dip=tilt=0
    vert_head = np.dot(rotmat, vert_axis) * vertlen
    vert_xs, vert_ys, vert_zs = (0, vert_head[0]), (0, vert_head[1]), (0, vert_head[2])

    return((major_xs, major_ys, major_zs),
           (minor_xs, minor_ys, minor_zs),
           (vert_xs, vert_ys, vert_zs))

def azmdip(point_vector):
    '''Calculates the azimuth/dip to a point (length 3 vector [x,y,z]) from the origin'''
    # Unpack and get angles using atan2 to avoid sign issues
    x, y, z = point_vector
    azm = 90.0 - np.degrees(np.arctan2(y, x))
    dip = np.degrees(np.arctan2(z, np.sqrt(x * x + y * y)))
    return(azm, dip)

def principaldirs(azm, dip, tilt):
    '''Calculate the principal directions from a plane of major continuity'''
    major, minor, vert = principalvectors(azm, dip, tilt)
    major_azm, major_dip = azmdip([major[0][1], major[1][1], major[2][1]])
    minor_azm, minor_dip = azmdip([minor[0][1], minor[1][1], minor[2][1]])
    vert_azm, vert_dip = azmdip([vert[0][1], vert[1][1], vert[2][1]])
    return((major_azm, major_dip), (minor_azm, minor_dip), (vert_azm, vert_dip))

class Arrow3D(FancyArrowPatch):
    '''Arrow3D from http://stackoverflow.com/questions/11140163/python-matplotlib-plotting-a-3d-cube-a-sphere-and-a-vector'''
    def __init__(self, xs, ys, zs, *args, **kwargs):
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
        self._verts3d = xs, ys, zs

    def draw(self, renderer):
        xs3d, ys3d, zs3d = self._verts3d
        xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
        self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
        FancyArrowPatch.draw(self, renderer)

def plotprincipalvectors(axes, azm, dip, tilt, lefthandrule=True,
                         majorlen=1.0, minorlen=1.0, vertlen=1.0):
    '''Add principal direction unit arrows to plot'''
    import matplotlib.lines as mlines
    # Principal dirs
    major, minor, vert = principalvectors(azm, dip, tilt, lefthandrule,
                                          majorlen=majorlen,
                                          minorlen=minorlen,
                                          vertlen=vertlen)

    # Major
    axes.add_artist(Arrow3D(major[0], major[1], major[2], mutation_scale=20, lw=2,
                            arrowstyle="-|>", color="g"))
    # Minor
    axes.add_artist(Arrow3D(minor[0], minor[1], minor[2], mutation_scale=20, lw=2,
                            arrowstyle="-|>", color="r"))
    # Vert
    axes.add_artist(Arrow3D(vert[0], vert[1], vert[2], mutation_scale=20, lw=2,
                            arrowstyle="-|>", color="b"))

    # Legend parameters
    majorline = mlines.Line2D([], [], color='g', marker=None,
                              markersize=0, lw=2.5, label='Major')
    minorline = mlines.Line2D([], [], color='r', marker=None,
                              markersize=0, lw=2.5, label='Minor')
    vertline = mlines.Line2D([], [], color='b', marker=None,
                             markersize=0, lw=2.5, label='Vert')

    return([majorline, minorline, vertline])

def drawgsvectorwidget(legend=True, sidetext=True, majortitle=False):
    """Draws an interactive GSLIB vector widget in an IPython notebook.

    Warning:
        Must be run with either:

            %matplotlib inline or %matplotlib notebook activated

    .. codeauthor:: Jared Deutsch 2016-02-21"""
    import ipywidgets

    def view_gslib_axes(azm=0.0, dip=0.0, tilt=0.0):
        '''Some function that is called view gslib axes'''

        fig = plt.figure(figsize=(7, 7))
        ax = plt.subplot(111, projection='3d', aspect='equal')

        mlines = plotprincipalvectors(ax, azm, dip, tilt)

        ax.set_xlim((-1.3, 1.3))
        ax.set_ylim((-1.3, 1.3))
        ax.set_zlim((-1.3, 1.3))
        ax.set_xlabel('Easting - X')
        ax.set_ylabel('Northing - Y')
        ax.set_zlabel('Elevation - Z')

        if legend:
            ax.legend(handles=mlines, frameon=False, loc='best')

        if sidetext:
            (major_azm, major_dip), (minor_azm, minor_dip), (vert_azm, vert_dip) = principaldirs(azm, dip, tilt)
            angles = ('Plane of major continuity (azm,dip,tilt) = {:.1f}, {:.1f}, {:.1f}\n\n'
                      'Major (azm,dip) = {:.1f}, {:.1f}\n'
                      'Minor (azm,dip) = {:.1f}, {:.1f}\n'
                      'Vert (azm,dip) = {:.1f}, {:.1f}\n').format(azm, dip, tilt, major_azm,
                                                                  major_dip, minor_azm, minor_dip,
                                                                  vert_azm, vert_dip)

            fig.text(1.0, 0.5, angles, fontsize='large')

        if majortitle:
            ax.set_title('azm = {} ({}), dip = {} ({}), tilt = {}'.format(azm, major_azm, dip,
                                                                          major_dip, tilt))
        plt.show()
        plt.close()
    ipywidgets.interact(view_gslib_axes,
                        azm=(-360.0, 360.0, 1.0),
                        dip=(-180.0, 180.0, 1.0),
                        tilt=(-90.0, 90.0, 1.0))

def drawellipsoid(ax, hmax, hmin, vert, azm=0.0, dip=0.0, tilt=0.0,
                  color='#87CEFA', alpha=0.5):
    """Draws an orientatable ellipsoid with GSLIB conventions.

    Spherical coordinate calculation from:

        http://stackoverflow.com/questions/7819498/plotting-ellipsoid-with-matplotlib

    Rotation uses GSLIB rotation matrix definition

    .. codeauthor:: Jared Deutsch 2016-03-06"""
    # Coefficients
    rx = hmax
    ry = hmin
    rz = vert

    # Set of all spherical angles:
    u = np.linspace(0, 2 * np.pi, 100)
    v = np.linspace(0, np.pi, 100)

    # Cartesian coordinates that correspond to the spherical angles:
    # (this is the equation of an ellipsoid):
    x = rx * np.outer(np.cos(u), np.sin(v))
    y = ry * np.outer(np.sin(u), np.sin(v))
    z = rz * np.outer(np.ones_like(u), np.cos(v))

    # Rotation matrix - uses the transpose of the GSLIB rotation matrix
    rotmat = get_rotation_matrix(azm, dip, tilt).transpose()

    # Flatten and rotate, then reshape
    xyzflat = np.array([x.reshape((100*100)),
                        y.reshape((100*100)),
                        z.reshape((100*100))])
    x, y, z = np.dot(rotmat, xyzflat)
    x = x.reshape((100, 100))
    y = y.reshape((100, 100))
    z = z.reshape((100, 100))

    ax.plot_surface(x, y, z, color=color, alpha=alpha,
                    rstride=4, cstride=4, lw=0.25,
                    antialiased=True)


def drawgsaniswidget(legend=True, sidetext=True, majortitle=False):
    """Draws an interactive anisotropy widget in an IPython notebook.

    Warning:
        Must be run with either:

            %matplotlib inline or %matplotlib notebook activated

    .. codeauthor:: Jared Deutsch 2016-03-06"""
    import ipywidgets

    def view_gslib_axes(azm=0.0, dip=0.0, tilt=0.0,
                        hmax=30.0, hmin=10.0, vert=5.0):
        '''Some function that is called view gslib axes'''

        fig = plt.figure(figsize=(9, 7))
        ax = plt.subplot(111, projection='3d', aspect='equal')

        mlines = plotprincipalvectors(ax, azm, dip, tilt,
                                      majorlen=1.3*max(hmax, hmin, vert),
                                      minorlen=1.3*max(hmax, hmin, vert),
                                      vertlen=1.3*max(hmax, hmin, vert))

        drawellipsoid(ax, hmax, hmin, vert, azm, dip, tilt)

        plotlim = max(ax.get_xlim()[1], ax.get_ylim()[1], ax.get_zlim()[1])
        ax.set_xlim((-1.3 * plotlim, 1.3 * plotlim))
        ax.set_ylim((-1.3 * plotlim, 1.3 * plotlim))
        ax.set_zlim((-1.3 * plotlim, 1.3 * plotlim))
        ax.set_xlabel('Easting - X')
        ax.set_ylabel('Northing - Y')
        ax.set_zlabel('Elevation - Z')

        if legend:
            ax.legend(handles=mlines, frameon=False, loc='best')

        if sidetext:
            (major_azm, major_dip), (minor_azm, minor_dip), (vert_azm, vert_dip) = principaldirs(azm, dip, tilt)
            angles = ('Plane of major continuity (azm,dip,tilt) = {:.1f}, {:.1f}, {:.1f}\n\n'
                      'Major (azm,dip) = {:.1f}, {:.1f}\n'
                      'Minor (azm,dip) = {:.1f}, {:.1f}\n'
                      'Vert (azm,dip) = {:.1f}, {:.1f}\n').format(azm, dip, tilt, major_azm,
                                                                  major_dip, minor_azm, minor_dip,
                                                                  vert_azm, vert_dip)

            fig.text(0.9, 0.5, angles, fontsize='large')

        if majortitle:
            ax.set_title('azm = {} ({}), dip = {} ({}), tilt = {}'.format(azm, major_azm, dip, major_dip, tilt))
        plt.show()
        plt.close()

    ipywidgets.interact(view_gslib_axes,
                        azm=(-360.0, 360.0, 1.0),
                        dip=(-180.0, 180.0, 1.0),
                        tilt=(-90.0, 90.0, 1.0),
                        hmax=(5.0, 1000.0, 5.0),
                        hmin=(5.0, 1000.0, 5.0),
                        vert=(5.0, 1000.0, 5.0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""plotting a 2D gridded dataset or a slice of a 3D gridded dataset with the ability to include another layer to show point data"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------
from . set_style import set_plot_style
import numpy as np
import matplotlib as mpl

@set_plot_style
def slice_plot(data, griddef=None, var=None, catdata=None, pointdata=None, pointvar=None,
			 pointtol=None, pointkws=None, pointcmap=None, orient='xy', slice_number=0, ax=None,
			 figsize=None, vlim=None, clim=None, title=None, xlabel=None, ylabel=None, unit=None,
			 rotateticks=None, cbar=True, cbar_label=None, catdict=None, cbar_label_pad=None,
			 cax=None, sigfigs=3, cmap=None, interp='none', aspect=None, grid=None, axis_xy=None,
			 rasterize=False, plot_style=None, custom_style=None, output_file=None, out_kws=None,
			 return_cbar=False, return_plot=False, slice_thickness=None, cbar_nticks=5,
			 plotformat_dict=None, **kwargs):
	"""
	slice_plot displays a 2D gridded dataset or a slice of a 3D gridded dataset. To plot only
	scattered data, please see :func:`gs.location_plot()<pygeostat.plotting.location_plot>`

	The only required parameters are ``data`` and ``griddef``. All other parameters are optional or
	calculated automatically. Axis tick labels are automatically checked for overlap and if needed,
	are rotated. The figure instance is always returned. To allow for easier modification, the
	colorbar object and the data used to generate the plot can also be returned. Examples of their
	use are provided bellow.

	The values use to bound the data (i.e., vmin and vmax) are automatically calculated by default.
	These values are determined based on the number of significant figures and the sliced data;
	depending on data and the precision specified, scientific notation may be used for the colorbar
	tick lables. When point data shares the same colormap as the gridded data, the points displayed
	are integrated into the above calculation.

	Categorical data can be used, however ``catdata`` will need to be set to ``True`` for proper
	implementation. Categorical colour palettes are available within pygeostat. See the
	documentation for :func:`gs.get_palette() <pygeostat.plotting.utils.get_palette>` for more
	information.

	Please review the documentation of the :func:`gs.set_style()
	<pygeostat.PlotStyle.set_style>` and :func:`gs.export_image()
	<pygeostat.plotting.export_image.export_image>` functions for details on their parameters so that
	their use in this function can be understood.

	Parameters:
		data: A numpy ndarray, pandas DataFrame or pygeostat DataFile, where each column is a
			variable and each row is an observation
		griddef (GridDef): A pygeostat GridDef class, which must be provided if a
			DataFile is not passed as data with a valid internal GridDef
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`
		var (str,int): The name of the column within data to plot. If an int is provided, then it
			corresponds with the column number in data. If None, the first column of data
			is used.
		catdata (bool): Indicate if the data is categorical or not. Will be automatically set if
			less than Parameters['plotting.assumecat'] unique values are found within ``data``
		pointdata (DataFile): A pygeostat DataFile class created using
			:class:`gs.DataFile <pygeostat.data.data.DataFile>` containg point data to overlay the
			gridded data with. Must have the necessary coordinate column headers stored within the
			class.
		pointvar (str): Column header of variable to plot within ``pointdata`` or a permissible
			matplotlib colour
		pointtol (float): Slice tolerance to plot point data (i.e. plot +/- ``pointtol`` from the
			center of the slice). Any negative value plots all data. Default is to plot all data.
		pointkws (dict): Optional dictionary of permissible keyword arguments to pass to
			matplotlib's scatter function. Default values is ``{'marker':'o', 's':15}``
		orient (str): Orientation to slice data. ``'xy'``, ``'xz'``, ``'yz'`` are the only accepted
			values
		slice_number (int): Grid cell location along the axis not plotted to take the slice of data to
			plot
		ax (mpl.axis): Matplotlib axis to plot the figure
		figsize (tuple): Figure size (width, height)
		vlim (float tuple): Data minimum and maximum values
		clim (int tuple) or (list): Categorical data minimum and maximum values, Forces categorical
			colorbar to plot the full range of categorical values - even if none show in the plot.
			Can be either a tuple of integer values OR a list of integers.
		title (str): Title for the plot. If left to it's default value of ``None`` or is set to
			``True``, a logical default title will be generated for 3-D data. Set to ``False`` if
			no title is desired.
		xlabel (str): X-axis label
		ylabel (str): Y-axis label
		unit (str): Unit to place inside the axis label parentheses
		rotateticks (bool tuple): Indicate if the axis tick labels should be rotated (x, y)
		cbar (bool): Indicate if a colorbar should be plotted or not
		cbar_label (str): Colorbar title
		catdict (dict): Dictionary to map enumerated catagories to names (e.g., {100: 'Unit1'}).
			Taken from Parameters['data.catdict'] if catdata=True and its keys align.
		sigfigs (int): Number of sigfigs to consider for the colorbar
		cmap (str): Matplotlib or pygeostat colormap or palette
		interp (str): plt.imshow interpolation option; ``'spline36'`` (continuous) and
			``'hermite'`` (categorical) are good starting points if smoothing is desired.
			``'none'`` is the default setting
		aspect (str): Set a permissible aspect ratio of the image to pass to matplotlib. If None,
			it will be 'equal' if each axis is within 1/5 of the length of the other. Otherwise,
			it will be 'auto'.
		grid (bool): Plots the major grid lines if True. Based on Parameters['plotting.grid']
			if None.
		axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
			visible) if axis_xy is True. Based on Parameters['plotting.axis_xy'] if None.
		rasterize (bool): Indicate if the gridded image should be rasterized during export. The
			output resolution is depended on the DPI setting used during export.
		plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
			:class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
			to turn it off
		custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
		output_file (str): Output figure file name and location
		out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
			:func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
		return_cbar (bool): Indicate if the colorbar axis should be returned. A tuple is returned
			with the first item being the axis object and the second the cbar object.
		return_plot (bool): Indicate if the plot from imshow should be returned. It can be used
			to create the colorbars required for subplotting with the ImageGrid()
		slice_thickness (int): Number of slices around the selected slice_number to average the attributes
		**kwargs: Optional permissible keyword arguments to pass to matplotlib's imshow function

	Returns:
		Default:
		ax (ax): Matplotlib axis instance which contains the gridded figure
		return_cbar:
		ax (ax), cbar (cbar): Optional, default False. Matplotlib colorbar object
		retrun_plot:
		ax (ax), plot(?): Optional, default False. Matplotlib colorbar object
		return_cbar & return_plot
		ax, plot, cbar: default False

	**Examples:**

		A simple call by passing the data file

		.. plot::

			import pygeostat as gs
			griddef = gs.GridDef([40,0.5,1,40,0.5,1,40,0.5,1])
			data_file = gs.ExampleData('3d_grid', griddef)
			gs.slice_plot(data_file, orient='xy', cmap='viridis')

		|

		Plot with slice thickness averaging/upscaling:

		.. plot::

			import pygeostat as gs
			griddef = gs.GridDef([40,0.5,1,40,0.5,1,40,0.5,1])
			data_file = gs.ExampleData('3d_grid', griddef)
			gs.slice_plot(data_file, orient='xy', cmap='viridis', slice_thickness=20)


	.. codeauthor:: pygeostat development team 2016-04-11
	"""

	import matplotlib.pyplot as plt
	from . utils import (_spatial_labels, _spatial_pointdata, _spatial_griddata, _spatial_slice,
						 _format_tick_labels, format_plot, _spatial_aspect, setup_plot, format_subplot_axis)

	from .. pygeostat_parameters import Parameters
	from .export_image import export_image
	from ..datautils.utils import slice_grid
	from ..data.data import DataFile


	# Handle dictionary defaults
	if out_kws is None:
		out_kws = dict()
	# Handle var
	if var is None:
		if isinstance(data, DataFile):
			if data.variables is not None:
				if isinstance(data.variables, str):
					var = data.variables
	# Handle pointvar
	if pointdata is not None:
		if pointvar is None:
			if isinstance(var, str):
				if var in list(pointdata.columns):
					pointvar = var
			if pointvar is None:
				pointvar = Parameters['plotting.location_plot.c']
	# Handle title
	if title is None:
		if hasattr(data, 'columns'):
			if var in data.columns:
				title = var
	# Parse the data, var and griddef input to determine the data and griddef
	data, griddef = _spatial_griddata(data, var, griddef)
	# some threshold of uniqueness means its likely categorical?
	# also the number of max categories in the palettes
	nunique = len(np.unique(data)[np.isfinite(np.unique(data))])
	if catdata is None and nunique < Parameters['plotting.assumecat']:
		catdata = True
	elif catdata is None and isinstance(catdict, dict):
		catdata = True
	elif catdata is None:  # back to original default argument
		catdata = False
	# Slice the data
	if orient in ['xy', 'xz', 'yz']:
		view = slice_grid(data, griddef, orient, slice_number, slice_thickness=slice_thickness)
	else:
		raise ValueError("Not a valid orientation! {}".format(orient))
	# Validate then slice the point data if required
	try:
		if pointtol < 0:
			pointtol = None
	except:
		pass
	if pointdata is not None:
		# Convert the pointdata into the data frame, x, y and z
		pointdata, pointx, pointy, pointz, _ = _spatial_pointdata(pointdata, orient)
		# Slice the data
		pointx, pointy, pointvari = _spatial_slice(pointdata, pointvar, pointx, pointy,
												   pointz, griddef, orient, slice_number, pointtol)
		if pointx is not None:
			if pointvar in pointdata.columns:
				if pointcmap is None:
					pointcmap = True
		else:
			pointcmap = None
	else:
		pointvari = None
		pointcmap = None
	# Handel grid extent
	a = orient[0]
	b = orient[1]
	xmin = getattr(griddef, a + 'limits')[0]
	xmax = getattr(griddef, a + 'limits')[1]
	ymin = getattr(griddef, b + 'limits')[0]
	ymax = getattr(griddef, b + 'limits')[1]
	if aspect is None:
		aspect = _spatial_aspect([xmin, xmax], [ymin, ymax])
	# handle parms for colorbars and colormaps
	output = _color_handling_gridded(data, cmap=cmap, catdata=catdata,
									 pointcmap=pointcmap, pointvar=pointvari,
									 view=view, sigfigs=sigfigs, vlim=vlim, clim=clim,
									 nticks=cbar_nticks, catdict=catdict)
	ticklabels, vlim, ticklocs, ncat, pointvari, cmap = output
	# Make sure the chosen colormap is set to the point data if required
	if pointcmap is True or pointcmap is None:
		pointcmap = cmap
	# Handel categorical data
	if catdata:
		tview = view.copy()
		for i in range(ncat):
			tview[view == ticklabels[i]] = i
		if catdict is None:
			catdict = Parameters['data.catdict']
		if isinstance(catdict, dict):
			ticklabels = [catdict[cat] for cat in ticklabels]
	else:
		tview = view.copy()
	# Set-up plot if no axis is supplied using the ImageGrid method if required or the regular way
	fig, ax, cax = setup_plot(ax, cbar=cbar, cax=cax, figsize=figsize)
	# Handle interpolation in aspect has changed
	if aspect is not 'equal' and interp is 'none':
		if catdata:
			interp = 'hermite'
		else:
			interp = 'spline36'
	# Plot gridded data
	plot = ax.imshow(tview, origin='lower', extent=[xmin, xmax, ymin, ymax], aspect=aspect,
					 interpolation=interp, vmin=vlim[0], vmax=vlim[1], cmap=cmap, **kwargs)
	# Rasterize the gridded plot
	if rasterize:
		plot.set_rasterized(True)
	# Plot scattered data if required
	if pointvari is not None:
		# Handel default dictionary
		if pointkws is None:
			pointkws = {'marker': 'o', 's': Parameters['plotting.location_plot.s']}
		if 'marker' not in pointkws:
			pointkws['marker'] = 'o'
		if 's' not in pointkws:
			pointkws['s'] = Parameters['plotting.location_plot.s']
		# if 'edgecolors' not in pointkws:
		#     pointkws['edgecolors'] = 'k'
		ax.scatter(pointx, pointy, c=pointvari, cmap=pointcmap,
				   vmin=vlim[0], vmax=vlim[1], **pointkws)
		ax.set_xlim((xmin, xmax))
		ax.set_ylim((ymin, ymax))
	# Plot labels
	ax = _spatial_labels(ax, orient, griddef, slice_number=slice_number, title=title, xlabel=xlabel,
						 ylabel=ylabel, unit=unit)
	# Note on Tick Labels:
	#   If a group of subplots are put together which share a x-axis, the rotation may not work. By
	#   getting the tick labels generated by matplotlib as a set of label objects, they can be
	#   looped through and have their settings individually fixed. This appears to be the only way
	#   to have the shared axis labels formated properly. The labels are also adjusted closer to
	#   the axis for esthetics. --Warren E. Black

	# The plots tick labels will not be properly accessible until the figure is "drawn", once the
	# command below is run, ax.get_ticklabel() will actually work properly.
	plt.draw()
	# Check to see if the ytick labels need to be rotated if the rotate argument was not passed.
	# This doesn't work if the figure is being placed into a subplot (i.e., is not a standalone
	# figure). Don't know why...
	_format_tick_labels(ax, rotateticks=rotateticks)
	# Note on Colorbar:
	#   It is my opinion that colorbars are more esthetically pleasing if they extend the full
	#   height of the figure it represents. The method that shifts the plot to the left distorts
	#   its aspect. A divider is used to accomplish the desired effect without distorting the
	#   figure. It is also important that the tick labels indicate that - in the case of a
	#   continuous variable - the min and max include their respective greater/less than or equal
	#    to symbols if they are not the absolute max/min values. --Warren E. Black
	# Plot colorbar, the tick locations and labels were generated above using gs.get_vlimdata
	if cbar:
		# Plot the colorbar
		cbar = fig.colorbar(plot, cax=cax, ticks=ticklocs)
		# Configure the color bar
		if cbar_label_pad is None:
			cbar_label_pad = 2
			for lbl in ticklabels:
				if isinstance(lbl, str):
					if 'geq' in lbl or 'leq' in lbl:
						cbar_label_pad = -6 + sigfigs
		cbar.ax.set_yticklabels(ticklabels, ha='left')
		cbar.ax.tick_params(axis='y', pad=2)
		if cbar_label is not None:
			cbar.set_label(cbar_label, ha='center', va='top', labelpad=cbar_label_pad)

	# Format plot based on plotformat_dict if it's available
	if plotformat_dict and cbar:
		fig, ax, cbar = format_subplot_axis(fig, ax, cbar=cbar,
											   plotformat_dict=plotformat_dict)
	elif plotformat_dict:
		fig, ax = format_subplot_axis(fig, ax, cbar=None,
										 plotformat_dict=plotformat_dict)
	if axis_xy is None:
		axis_xy = Parameters['plotting.axis_xy_spatial']
	format_plot(ax, grid=grid, axis_xy=axis_xy)

	# Export figure
	if output_file or ('pdfpages' in out_kws and out_kws['pdfpages']):
		export_image(output_file, **out_kws)
	# Return whats needed
	if return_cbar:
		if return_plot:
			return ax, plot, cbar
		else:
			return ax, cbar
	elif return_plot:
		return ax, plot
	else:
		return ax


def _color_handling_gridded(data, cmap=None, catdata=None, pointcmap=None,
							pointvar=None, view=None, sigfigs=3, vlim=None,
							clim=None, ticklabels=None, ticklocs=None, nticks=5,
							catdict=None):
	'''A small utility function called from many of the plotting functions. This sets the colorbar
	and colormap parameters and loads the specified colormap.

	Parameters:
		data: Tidy (long-form) dataframe where each column is a variable and each row is an
			observation. Pandas dataframe or numpy array
		cmap (str): Matplotlib or pygeostat colormap or palette
		catdata (bool): Indicate if the data is categorical or not. Will be automatically set if
			less than Parameters['plotting.assumecat'] unique values are found within ``data``
		pointcmap (bool): True to color the points using cmap
		pointvar: a pandas series of the point data used for the colormap. If categorical data is
			passed this will get converted to sequential numbers used in the colormap object.
		view: A slice of the data using ``gs.slice_grid()``
		sigfigs (int): Number of sigfigs to consider for the colorbar
		vlim (float tuple): Data minimum and maximum values
		clim (int tuple) or (list): Categorical data minimum and maximum values, Forces categorical
			colorbar to plot the full range of categorical values - even if none show in the plot.
			Can be either a tuple of integer values OR a list of integers.
		ticklabels: Filler. Otherwise you seem to have to set this variable before calling the
			function otherwise it gives you an error about assigning a value before the variable is
			defined
		ticklocs: Filler. Otherwise you seem to have to set this variable before calling the
			function otherwise it gives you an error about assigning a value before the variable is
			defined
		nticks: the number of ticks on the colorbar

	Returns:
		ticklabels: Matplotlib ticklabels
		vlim: If catdata is passed then this is reset to match the colormap. None otherwise
		ticklocs: Matplotlib ticklocations
		ncat: if categorical data is found then this is the number of categories. None otherwise
		pointvar: If categorical data is found then the point view is mapped to the colormap values
		cmap: a matplotlib colormap object
		'''
	from . utils import _get_cmap, get_contcbarargs

	if catdata:
		if clim:
			if isinstance(clim, tuple):
				ticklabels = np.arange(clim[0], (clim[1] + 1))
			elif isinstance(clim, list):
				ticklabels = [int(x) for x in clim]
			else:
				raise TypeError('Invalid clim type. Must be tuple or list')
		else:
			ticklabels = np.unique(data[~np.isnan(data.astype(float))])
			ticklabels = [int(x) for x in ticklabels]
		ncat = len(ticklabels)
		if isinstance(catdict, dict) and ncat != len(catdict):
			# modify the cmap to store the colors as if cmap is generated from all cats
			cmap = _get_cmap(cmap, catdata, len(catdict))
			cmap = mpl.colors.ListedColormap(
				[clr for cat, clr in zip(catdict, cmap.colors) if cat in ticklabels]
			)
		else:
			cmap = _get_cmap(cmap, catdata, ncat)
		vlim = (0, ncat)
		ticklocs = np.arange(ncat) + 0.5
		if pointcmap is True:
			dump = pointvar.copy()
			for i in range(ncat):
				dump[pointvar == ticklabels[i]] = i
			pointvar = dump
	else:
		if (pointvar is not None) and (pointcmap is not None):
			try:
				temp_pointdat = pointvar.values()
			except:
				temp_pointdat = pointvar
			temp = np.append(temp_pointdat, view)
			vlim, ticklocs, ticklabels = get_contcbarargs(temp, sigfigs, vlim,
															 nticks=nticks)
		else:
			vlim, ticklocs, ticklabels = get_contcbarargs(view, sigfigs, vlim,
															 nticks=nticks)
		ncat = None
		cmap = _get_cmap(cmap, catdata, ncat)
	return ticklabels, vlim, ticklocs, ncat, pointvar, cmap



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""A basic quantile by quantile plot to compare two probability distribution"""
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------


from . set_style import set_plot_style


@set_plot_style
def qq_plot(data, reference_data, data_weight=None, reference_weight=None, limits=None, npoints=0,
          log_scale=None, ax=None, figsize=None, title=None, xlabel=None,
          ylabel=None, s=None, percent=True, color='k', grid=None, axis_xy=None, plot_style=None,
          custom_style=None, output_file=None, out_kws=None, line=True, ntickbins=5, **kwargs):
    """
    Plot a QQ plot between the reference. Pretty much the probplt but with 2 datasets and plotting
    the quantiles between them

    Parameters:
        data: Tidy (long-form) 1D data where a single column of the variable exists with each row
            is an observation. A pandas dataframe/series or numpy array can be passed.
        reference_data: Tidy (long-form) 1D data or a valid scipy.stats distribution (e.g. "norm"). A
            pandas dataframe/series or numpy array can be passed.
        data_weight: 1D dataframe, series, or numpy array of declustering weights for the data.
        reference_weight: 1D dataframe, series, or numpy array of declustering weights for the data.
        lower (float): Lower trimming limits
        upper (float): Upper trimming limits
        limits (tuple): the min and max value of the axes
        ax (mpl.axis): Matplotlib axis to plot the figure
        log_scale (bool): yes or no to log_scale
        npoints (int): set to 0 to use all points
        figsize (tuple): Figure size (width, height)
        xlim (float tuple): Minimum and maximum limits of data along the x axis
        title (str): Title for the plot
        xlabel (str): X-axis label. A default value of ``None`` will try and grab a label from the
            passed ``data``. Pass ``False`` to not have an xlabel.
        s (int): Size of points
        color (str or int): Any permissible matplotlib color or a integer which is used to draw a
            color from the pygeostat color pallet ``pallet_pastel`` (useful for iteration)
        grid(bool): plots the major grid lines if True. Based on gsParams['plotting.grid']
            if None.
        axis_xy (bool): converts the axis to GSLIB-style axis visibility (only left and bottom
            visible) if axis_xy is True. Based on gsParams['plotting.axis_xy'] if None.
        plot_style (str): Use a predefined set of matplotlib plotting parameters as specified by
            :class:`gs.GridDef <pygeostat.data.grid_definition.GridDef>`. Use ``False`` or ``None``
            to turn it off
        custom_style (dict): Alter some of the predefined parameters in the ``plot_style`` selected.
        output_file (str): Output figure file name and location
        out_kws (dict): Optional dictionary of permissible keyword arguments to pass to
            :func:`gs.export_image() <pygeostat.plotting.export_image.export_image>`
        line (bool): Plot the reference 1:1 line
        ntickbins (int or tuple): modify the number of ticks. Only works if log_scale == ``True``
        **kwargs: Optional permissible keyword arguments to pass to matplotlib's scatter function

    Returns:
        ax (ax): matplotlib Axes object with the histogram

    **Examples:**

    A simple call:
    
    .. plot::

        import pygeostat as gs
        import numpy as np
        # load some data
        gs.qq_plot(np.random.randn(1000),np.random.randn(1000))

    """

    # Import the rest of the packages
    import numpy as np
    import pandas as pd
    from .utils import get_label, format_plot
    from .export_image import export_image
    from ..statistics.cdf import cdf
    import matplotlib.pyplot as plt

    if xlabel is None:
        xlabel = get_label(data)
    if ylabel is None:
        ylabel = get_label(reference_data)
    # Coerce the passed data and wt into a single pandas dataframe
    
    try:
        data = pd.DataFrame(data=data)
        data.reset_index(inplace=True, drop=True)
        if data.shape[1] != 1:
            raise ValueError("The passed `data` is not 1-D")
    except:
        raise ValueError("Please ensure the passed `data` can be coerced into a pandas dataframe"
                         " (i.e., data = pd.DataFrame(data=data)")
    if data_weight is not None:
        try:
            data_weight = pd.DataFrame(data=data_weight, columns=["wt"])
            if data_weight.shape[1] != 1:
                raise ValueError("The passed `data_weight` is not 1-D")
        except:
            raise ValueError("Please ensure the passed `data_weight` can be coerced into a pandas"
                             " dataframe (i.e., data_weight = pd.DataFrame(wt=data_weight)")
    else:
        data_weight = pd.DataFrame(data=np.ones((len(data), 1)))
    data = pd.concat([data, data_weight], axis=1)
    data.columns = ['data', 'data_weight']

    # Do the same for the reference data
    try:
        reference_data = pd.DataFrame(data=reference_data)
        reference_data.reset_index(inplace=True, drop=True)
        if reference_data.shape[1] != 1:
            raise ValueError("The passed `reference_data` is not 1-D")
    except:
        raise ValueError("Please ensure the passed `reference_data` can be coerced into a pandas dataframe"
                         " (i.e., reference_data = pd.DataFrame(data=reference_data)")
    if reference_weight is not None:
        try:
            reference_weight = pd.DataFrame(data=reference_weight)
            if reference_weight.shape[1] != 1:
                raise ValueError("The passed `reference_weight` is not 1-D")
        except:
            raise ValueError("Please ensure the passed `reference_weight` can be coerced into a pandas"
                             " dataframe (i.e., reference_weight = pd.DataFrame(reference_weight=reference_weight)")
    else:
        reference_weight = pd.DataFrame(data=np.ones((len(reference_data), 1)))
    reference_data = pd.concat([reference_data, reference_weight], axis=1)
    reference_data.columns = ['reference_data', 'reference_weight']

    # Handle dictionary defaults
    if out_kws is None:
        out_kws = dict()

    # build the cdf's using all bins
    print(len(data))
    mp1, p1 = cdf(data['data'], weights=data['data_weight'])
    mp2, p2 = cdf(reference_data['reference_data'], weights=reference_data['reference_weight'])

    if npoints <= 0:
        npoints = min([len(data), len(reference_data), 1e10])

    # generate and interpolate the equally spaced set of quantiles
    probs = np.linspace(1 / npoints, 1, npoints)
    x = np.interp(probs, p2, mp2)
    y = np.interp(probs, p1, mp1)

    # Set-up plot if no axis is supplied
    if ax is None:
        _, ax = plt.subplots(figsize=figsize)
    # Plot the figure
    ax.scatter(x, y, s=s, c=color, lw=0, **kwargs)

    if limits is None:
        limits = np.minimum(np.min(mp1), np.min(mp2)), np.maximum(np.max(mp1), np.max(mp2))

    # apply the log_scale
    if log_scale:
        ax.set_xscale('log')
        ax.set_yscale('log')

    if line:
        ax.plot(limits, limits, color='r', lw=0.5)

    # deal with the labels and formatting
    if xlabel is None:
        xlabel = '$ref_{quantiles}$'
    if ylabel is None:
        ylabel = '$data_{quantiles}$'
    ax = format_plot(ax, xlabel=xlabel, ylabel=ylabel, title=title, grid=grid, axis_xy=axis_xy,
                   xlim=limits, ylim=limits)
    # Export figure
    if output_file or ('pdfpages' in out_kws):
        export_image(output_file, **out_kws)

    return ax



'''pygeostat: Python interface for geostatistics
 Copyright (C) 2013-2020 pygeostat core team'''

import os
import subprocess
import setuptools
from setuptools import setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open ('requirements.txt', 'r') as file:
    install_requires = file.readlines()
install_requires = [x.strip() for x in install_requires]


setup(name='pygeostat',
      version='1.0.5',
      description='Python interface for geostatistics',
      long_description = long_description,
      long_description_content_type="text/x-rst",
      url='http://www.ccgalberta.com/pygeostat/welcome.html',
      project_urls ={
          'Documentation': 'http://www.ccgalberta.com/pygeostat/welcome.html'
          ,'Source': 'https://github.com/CcgAlberta/pygeostat'
      },
      maintainer='pygeostat development team',
      maintainer_email='hadavand@ualberta.ca',
      author='Jared Deutsch, Matthew Deutsch, Ryan Martin, Warren Black, Tyler Acorn, Mostafa Hadavand, Ryan Barnett',
      license='MIT License',
      packages=setuptools.find_packages(),
      python_requires='>3.5',
      install_requires = install_requires,
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License", 
        "Operating System :: Microsoft", 
       ],
      package_data={'pygeostat': ['data/example_data/*', 'executable/*.txt']},
      include_package_data=True,
      zip_safe=False)


# To install, run python with the following command line:
# >>> pip install .
# Alternative installation, run python with the following command line:
# python setup.py install
# To uninstall, run the command:
# >>> pip uninstall pygeostat
# To create distribution files
# >>> python setup.py bdist_wheel


#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
iotools.py: Contains input/output utilities/functions for pygeostat. Many of which
are based off of Pandas builtin functions.
'''
#-----------------------------------------------------------------------------
# Boilerplate
#-----------------------------------------------------------------------------
from __future__ import absolute_import, division, print_function, unicode_literals

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import warnings

import pandas as pd
import numpy as np
from .. pygeostat_parameters import Parameters

def read_file(flname, fltype=None, headeronly=False, delimiter=r'\s*', h5path=None, h5datasets=None,
             columns=None, ireal=1, griddef=None, tmin=None):
    '''
    Reads in a GSLIB-style Geo-EAS data file, CSV, or HDF5 data files.

    Parameters:
        flname (str): Path (or name) of file to read.

    Keyword Args:
        fltype (str): Type of file to read: either ``csv``, ``gslib``, or ``hdf5``.
        headeronly (bool): If True, only reads in the 1st line from the data file
                     which is useful for just getting column numbers or testing. OR
                     it allows you to open a hdf5 object with Pandas HDFStore functionality
        delimiter (str): Delimiter specified instead of sniffing
        h5path (str): Forward slash (/) delimited path through the group hierarchy you wish to
            read the dataset(s) specified by the argument ``datasets`` from. The dataset name
            cannot be passed using this argument, it is interpreted as a group name only. A
            value of ``None`` places the dataset into the root directory of the HDF5 file. A value
            of ``False`` loads a blank pd.DataFrame().
        h5datasets (str or list): Name of the dataset(s) to read from the group specified by
            ``h5path``. Does nothing if ``h5path`` points to a dataset.
        column (list): List of column labels to use for resulting frame
        ireal (int): Number of realizaitons in the file
        griddef (GridDef): griddef for the realization
        tmin (float): values less than this number are convernted to NaN, since NaN's are
            natural handled within matplotlib, pandas, numpy, etc. If None, set to
            pygeostat.Parameters['data.tmin'].

    Returns:
        data (pandas.DataFrame): Pandas DataFrame object with input data.

    Note:
        Functions can also be called seperately with the following code

        >>> data.data = pygeostat.read_gslib(flname)
        >>> data.data = pygeostat.read_csv(flname)
        >>> data.data = pygeostat.read_h5(flname, h5path='')
        >>> data.data = pygeostat.open_hdf5(flname)

    '''
    # Infer filetype if none specified based on the file extention
    if fltype is None:
        import os
        _, extention = os.path.splitext(flname)
        extention = extention.lower()
        if extention in ['.out', '.data']:
            fltype = 'gslib'
        elif extention in ['.csv']:
            fltype = 'csv'
        elif extention in ['.h5', '.hdf5', '.hd5']:
            fltype = 'hdf5'
        else:
            # Otherwise just assume GSLIB for now but this may change
            fltype = 'gslib'

    # Call specific read functions based on data type
    if fltype == 'gslib':
        data = read_gslib(flname=flname, headeronly=headeronly, delimiter=delimiter,
                          tmin=tmin)
    elif fltype == 'csv':
        data = read_csv(flname=flname, headeronly=headeronly, tmin=tmin)
    elif fltype == 'hdf5' or fltype == 'hd5' or fltype == 'h5':
        from .h5_io import read_h5
        if headeronly:
            data = pd.DataFrame()
        else:
            # Note that tmin isn't implemented for h5 yet, due to my inexperience
            # with this underlying module/format
            data = read_h5(flname=flname, h5path=h5path, datasets=h5datasets)
    else:
        print('File type unsupported! Try "gslib", "csv" or "hdf5"')

    if columns is not None:
        data.columns = columns

    return data


def _test_file_open(filename):
    """ test and raise if a file cannot be opened """
    try:
        with open(filename, 'r'):
            pass
    except FileNotFoundError:
        raise FileNotFoundError('{} does not exist!'.format(filename))
    except IOError:
        raise IOError('Could not open {}'.format(filename))


def read_gslib(flname, headeronly=False, delimiter=r'\s*', tmin=None):
    '''Reads in a GSLIB-style Geo-EAS data file

    Parameters:
        flname (str): Path (or name) of file to read.

    Keyword Args:
        headeronly (bool): If True, only reads in the 1st line from the data file
                     which is useful for just getting column numbers or testing
        delimiter (str): Delimiter specified instead of sniffing
        tmin (float): values less than this number are convernted to NaN, since NaN's are
            natural handled within matplotlib, pandas, numpy, etc. If None, set to
            pygeostat.Parameters['data.tmin'].

    Returns:
        data (pandas.DataFrame): Pandas DataFrame object with input data.

    '''
    # Can the file be opened?
    _test_file_open(flname)
    # Only read in the header + 1 line of the data file?
    if headeronly:
        nrows = 1
    else:
        nrows = None
    # Reading the header for a GSLIB "Geo-EAS" format file
    with open(flname, 'r') as datafl:
        # Geo-EAS Header
        _ = datafl.readline().strip()  # Title
        nvar = int(datafl.readline().split()[0])
        varnames = []
        for _ in range(nvar):
            varnames.append(datafl.readline().strip())
        nrowtoskip = 2 + nvar

    engine = Parameters['data.io.pandas_engine']
    # Use the new pandas chunked reading method
    # with the "python" engine since the C engine sometimes does not work with
    # realizations with trailing spaces
    try:
        tpdf = pd.read_csv(flname, skiprows=nrowtoskip, header=None,
                           delimiter=delimiter, skipinitialspace=True,
                           nrows=nrows, engine=engine, chunksize=100000)
        data = pd.concat(tpdf, ignore_index=True)
    # Fallback to old method using the python engine
    except (ValueError, NotImplementedError):
        data = pd.read_csv(flname, skiprows=nrowtoskip, header=None, delimiter=r'\s*',
                           skipinitialspace=True, nrows=nrows, engine='python')

    # Replace values below tmin with nan
    data = _data_trim(data, tmin=tmin)

    # Assign variable names
    data.columns = varnames

    # return only the Pandas DataFrame
    return data


def read_csv(flname, headeronly=False, tmin=None):
    '''Reads in a GSLIB-style CSV data file.

    Parameters:
        flname (str): Path (or name) of file to read.

    Keyword Args:
        headeronly (bool): If True, only reads in the 1st line from the data file
                     which is useful for just getting column numbers or testing
        delimiter (str): Delimiter specified instead of sniffing
        tmin (float): values less than this number are convernted to NaN, since NaN's are
            natural handled within matplotlib, pandas, numpy, etc. If None, set to
            pygeostat.Parameters['data.tmin'].

    Returns:
        data (pandas.DataFrame): Pandas DataFrame object with input data.

    '''
    if headeronly:
        nrows = 1
    else:
        nrows = None
    # Reading a GSLIB "Geo-EAS" format file
    tpdf = pd.read_csv(flname, header=0, nrows=nrows, engine='c', chunksize=100000)
    data = pd.concat(tpdf, ignore_index=True)
    # Replace values below tmin with nan
    data = _data_trim(data, tmin=tmin)
    # Only return the DataFrame
    return data


def _data_trim(data, tmin):
    """Replace values less than tmin with NaN. This routine is private
       to the iotools module, since it is only intended for use on the
       import of data.

    Parameters:
        data (pandas.DataFrame): data to trim
        tmin (int or float): values less than tmin are assigned np.nan

    """
    from pandas.api.types import is_numeric_dtype
    if tmin is None:
        # Use the Parameters tmin, which may also be None (leading to no trimming)
        tmin = Parameters['data.tmin']
    if tmin is None or not tmin:
        return data
    if not isinstance(tmin, int) and not isinstance(tmin, float):
        raise ValueError('tmin must be an integer or float!')
    if tmin is not None:
        if Parameters['data.fix_legacy_null'] and Parameters['data.null'] is not None:
            warned = False
            for legacy_null in Parameters['data.legacy_null']:
                if not warned and legacy_null != Parameters['data.null']:
                    for col in data.columns:
                        if is_numeric_dtype(data[col]):
                            if (np.isclose(data[col], legacy_null)).any():
                                warnings.warn(
                                    f'found {legacy_null} in {col} during read!')
                                warned = True
                                break
                data.replace(legacy_null, Parameters['data.null'], inplace=True)
        for col in data.columns:
            if is_numeric_dtype(data[col]):
                data.loc[data[col] < tmin, col] = np.nan
    return data


def write_gslib(data, flname, title=None, variables=None,
                fmt=None, sep=' ', null=None):
    """Writes out a GSLIB-style data file.

    Parameters:
        data (pygeostat.DataFile or pandas.DataFrame): data to write out
        flname (str): Path (or name) of file to write out.

    Keyword Args:
        title (str): Title for output file.
        variables (List(str)): List of variables to write out if only a subset is desired.
        fmt (str): Format to use for floating point numbers.
        sep (str): Delimiter to use for file output, generally don't need to change.
        null (float): NaN numbers are converted to this value prior to writing. If None, set
            to data.null. If data.Null is None, set to pygeostat.Parameters['data.null'].

    """
    from .data import DataFile as DataFile
    data = _data_fillnan(data, null)
    # If a DataFile is used, check the arguments
    if isinstance(data, DataFile):
        flname, variables, sep, fltype = data.check_datafile(flname, variables,
                                                             sep, 'gslib')
    else:
        # If variables is none, then get the columns
        if variables is None:
            variables = data.columns.tolist()
    # If title is STILL none after introspection, then at least include something
    if title is None:
        title = 'pygeostat_saved_data'
    if fmt is None:
        fmt = Parameters.get('data.write.python_floatfmt', '%.5f')
    # Force a write
    with open(flname, 'w') as outfl:
        # GSLIB Header
        outfl.write(title.strip() + '\n' + str(len(variables)) + '\n')
        for variable in variables:
            outfl.write(str(variable) + '\n')
        # GSLIB separated data
        if isinstance(data, DataFile):
            data.data[variables].to_csv(outfl, header=False, index=False, sep=sep,
                                        float_format=fmt, line_terminator='\n')
        else:
            data[variables].to_csv(outfl, header=False, index=False, sep=sep,
                                   float_format=fmt, line_terminator='\n')


def write_csv(data, flname, variables=None,
              fmt='%.5f', sep=',', fltype='csv', null=None):
    """Writes out a CSV or Excel (XLSX) data file.

    Parameters:
        data (pygeostat.DataFile or pandas.DataFrame): data to write out
        flname (str): Path (or name) of file to write out.

    Keyword Args:
        variables (List(str)): List of variables to write out if only a subset is desired.
        fmt (str): Format to use for floating point numbers.
        sep (str): Delimiter to use for file output, generally don't need to change.
        fltype (str): Type of file to write either ``csv`` or ``xlsx``.
        null (float): NaN numbers are converted to this value prior to writing. If None, set
            to data.null. If data.Null is None, set to pygeostat.Parameters['data.null'].

    """
    from .data import DataFile as DataFile
    data = _data_fillnan(data, null)
    # If a DataFile is used, check the arguments
    if isinstance(data, DataFile):
        flname, variables, sep, fltype = data.check_datafile(flname, variables,
                                                             sep, fltype)
    else:
        # If variables is none, then get the columns
        if variables is None:
            variables = data.columns.tolist()
    # Force a write
    with open(flname, 'w') as outfl:
        if fltype.lower() == 'csv':
            # CSV using pandas native CSV writer
            if isinstance(data, DataFile):
                data.data[variables].to_csv(outfl, header=True, index=False, sep=sep,
                                            float_format=fmt, line_terminator='\n')
            else:
                data[variables].to_csv(outfl, header=True, index=False, sep=sep,
                                       float_format=fmt, line_terminator='\n')
        elif (fltype.lower() == 'xlsx') or (fltype.lower() == 'excel'):
            # Excel file writer - only xlsx is supported naturally in pygeostat,
            # but XLWT could be used if XLS files needed to be created
            # NOTE: this requires the Python package openpyxl
            if isinstance(data, DataFile):
                data.data[variables].to_excel(flname, header=True, index=False,
                                              float_format=fmt)
            else:
                data[variables].to_excel(flname, header=True, index=False,
                                         float_format=fmt)


def write_vtk(data, flname, dftype=None, x=None, y=None, z=None, variables=None, griddef=None,
              null=None, vdtype=None, cdtype=None):
    """
    Writes out an XML VTK data file. A required dependency is pyevtk, which may be installed using
    the following command:

        >>> pip install pyevtk

    Users are also recommended to install the latest Paraview, as versions from 2017 were
    observed to have odd precision bugs with the XML format.

    Parameters:
        data (pygeostat.DataFile): data to write out
        flname (str): Path (or name) of file to write out (without extension)

    Keyword Args:
        dftype (str): type of datafile options ``grid`` or ``point``, which if None,
            is drawn from data.dftype
        x (str): name of the x-coordinate, which is used if ``point``. Drawn
            from data.x if the kwarg=None. If not provided by these means for ```sgrid```,
            calculated via ``sim.griddef.get_coordinates()``.
        y (str): name of the y-coordinate, which is used if ``point``. Drawn
            from data.y if the kwarg=None. If not provided by these means for ```sgrid```,
            calculated via sim.griddef.get_coordinates().
        z (str): name of the z-coordinate, which is used if ``point``. Drawn
            from data.z if the kwarg=None. If not provided by these means for ```sgrid```,
            calculated via sim.griddef.get_coordinates().
        griddef (pygeostat.GridDef): grid definition, which is required if ``grid``.
            Drawn from data.griddef if the kwarg=None.
        variables (list or str): List or string of variables to write out. If None, then all
            columns aside from coordinates are written out by default.
        null (float): NaNs are converted to this value prior to writing. If None, set to
            pygeostat.Parameters['data.null_vtk'].
        vdtype (dict(str)) : Dictionary of the format {'varname': dtype}, where dtype is
            a numpy data format. May be used for reducing file size, by
            specifying ``int``, ``float32``, etc. If a format string is provided instead
            of a dictionary, that format is applied to all variables. This is not applied
            to coordinate variables (if applicable). If None, the value is drawn from
            Parameters['data.write_vtk.vdtype'].
        cdtype (str) : Numpy format to use for the output of coordinates, where valid formats
            are ``float64`` (default) and ``float32``. The later is recommended for reducing
            file sizes, but may not provide the requisite precision for UTM coordinates. If None,
            the value is drawn from Parameters['data.write_vtk.cdtype'].

    ``dftype`` should be one of:

    1. 'point' (irregular points) where ``data.x``, ``data.y`` and ``data.z`` are columns in ``data.data``
    2. 'grid' (regular or rectilinear grid) where ``data.griddef`` must be initialized
    3. 'sgrid' (structured grid) where ``data.x``, ``data.y`` and ``data.z`` are columns in ``data.data``. ``data.griddef`` should also be initialized, although only ``griddef.nx``, ``griddef.ny`` and ``griddef.nz`` are utilized (since the grid is assumed to not be regular)

    """
    from copy import deepcopy
    # Ensure dependencies are satisfied
    try:
        from pyevtk.hl import pointsToVTK, gridToVTK
    except:
        raise Exception('Looks like pyevtk is not installed, use:\n' +
                        '>>> pip install pyevtk\n')
    # Place in a new variable so potential changes aren't returned
    dat1 = deepcopy(data)
    # Ensure valid VTK option
    if dftype is None:
        if dat1.dftype == 'point':
            dftype = 'point'
        elif dat1.dftype == 'grid':
            dftype = 'grid'
        elif dat1.dftype == 'sgrid':
            dftype = 'sgrid'
        else:
            raise ValueError(" specified data.dftype is not yet available or invalid;\n" +
                             " see the docstring!")
    else:
        if not any([dftype == test for test in ['point', 'grid', 'sgrid']]):
            raise ValueError(" specified dftype is not yet available or invalid;\n" +
                             " see the docstring!")
    # Ensure a grid is present if necessary
    if dftype in ['grid', 'sgrid']:
        if griddef is None:
            griddef = dat1.griddef
            if griddef is None:
                raise ValueError(('griddef must be a kwarg or exist as data.griddef if '
                                  'dftype is grid or sgrid!'))
    # Ensure coordinates are present if necessary
    if dftype in ['point', 'sgrid']:
        if x is None:
            x = dat1.x
        if y is None:
            y = dat1.y
        if z is None:
            z = dat1.z
        test = dict(x=x, y=y, z=z)
        if dftype == 'point':
            for coord in test.keys():
                if test[coord] is None:
                    raise ValueError(('{} must be a kwarg or exist as data.{} if'
                                      'dftype is point'.format(coord, coord)))
        else:
            t = [t is None for t in test.values()]
            if all(t):
                raise ValueError(('at least one grid coordinate should be irregular (specified '
                                  'as a column in data) if using dftype="sgrid". Specify one '
                                  'or multiple irregular coordinate columns, or use dftype="grid"'))
            else:
                # A coordinate is not provided, so assume it follows the regular grid
                tx, ty, tz = griddef.get_coordinates()
                if x is None:
                    x, dat1['X'] = 'X', tx
                if y is None:
                    y, dat1['Y'] = 'Y', ty
                if z is None:
                    z, dat1['Z'] = 'Z', tz
    # Replace nan with null values
    if null is None:
        null = Parameters['data.null_vtk']
    dat1 = _data_fillnan(dat1, null)
    # Check the grid length
    if dftype != 'point':
        nx = griddef.nx
        ny = griddef.ny
        nz = griddef.nz
        nxyz = nx * ny * nz
        if dat1.shape[0] != nxyz:
            raise ValueError('nx*ny*nz should be equal to dat1.shape[0]!')
    if vdtype is None:
        vdtype = Parameters['data.write_vtk.vdtype']
    if cdtype is None:
        cdtype = Parameters['data.write_vtk.cdtype']
    # Generate a dictionary of the variables, including their speficied precision
    columns = dat1.data.columns.tolist()
    if variables is None:
        variables = []
        for col in columns:
            if col == x or col == y or col == z:
                continue
            variables.append(col)
        if len(variables) == 0:
            # Paraview requires some dat1...
            variables = [x]
    elif isinstance(variables, str):
        variables = [variables]
    # Generate a dictionary of variables precisions
    vdtyped = {col: 'float64' for col in columns}
    if isinstance(vdtype, dict):
        for var in vdtype:
            vdtyped[var] = vdtype[var]
    elif isinstance(vdtype, str):
        for var in vdtyped:
            if var == x or var == y or var == var == z:
                vdtyped[var] = cdtype
            else:
                vdtyped[var] = vdtype
    vardict = dict()
    for var in variables:
        if any(var in i for i in columns):
            t = np.ascontiguousarray(dat1.data[var].values)
            vardict[var] = t.astype(vdtyped[var])
            if dftype != 'point':
                vardict[var] = vardict[var].reshape((nx, ny, nz), order='F')
        else:
            raise KeyError(var + " is not a column name in dat1.data!")
    # General checks and variable assembly completed - write out the
    # data type
    if dftype == 'point':
        # Require specified coodinates for each point
        if any(x in i for i in columns):
            x = np.ascontiguousarray(dat1.data[x].values)
            x = x.astype(cdtype)
        else:
            raise KeyError(" x is not a column name in data.data!")
        if any(y in i for i in columns):
            y = np.ascontiguousarray(dat1.data[y].values)
            y = y.astype(cdtype)
        else:
            raise KeyError(" y is not a column name in data.data!")
        if any(z in i for i in columns):
            z = np.ascontiguousarray(dat1.data[z].values)
            z = z.astype(cdtype)
        else:
            raise KeyError(" z is not a column name in data.data!")
    if dftype == 'point':
        pointsToVTK(flname, x, y, z, data=vardict)
    elif dftype == 'grid':
        # Rectilinear VTK type
        grid_arr = data.griddef.grid_array
        xsiz, ysiz, zsiz = grid_arr[2], grid_arr[5], grid_arr[8]
        xmin, ymin, zmin = grid_arr[1] - xsiz * .5, grid_arr[4] - \
            ysiz * .5, grid_arr[7] - zsiz * .5
        xmax, ymax, zmax = xmin + nx * xsiz, ymin + ny * ysiz, zmin + nz * zsiz
        x = np.arange(xmin, xmax + xsiz, xsiz, dtype=cdtype)
        y = np.arange(ymin, ymax + ysiz, ysiz, dtype=cdtype)
        z = np.arange(zmin, zmax + zsiz, zsiz, dtype=cdtype)
        gridToVTK(flname, x, y, z, cellData=vardict)
    else:
        pass


def _data_fillnan(data, null):
    """Replace nan values prior to writing out. Private since only intended
       for use within the iotools

    Parameters:
        data (pandas.DataFrame or pygeostat.DataFrame): data that may have NaN values
        null (int or float): NaN occurence are replaced with this value

    """
    import pygeostat as gs
    import copy
    if null is None:
        if isinstance(data, gs.DataFile):
            if data.null is None:
                # Use the Parameters null, which may also be None (leading to no nan assignment)
                null = Parameters['data.null']
            else:
                # Use the DataFile.null, which may also be None (leading to no nan assignment)
                null = data.null
        else:
            null = Parameters['data.null']
    if null is None or not null:
        return data
    if not isinstance(null, int) and not isinstance(null, float):
        raise ValueError('null must be an integer or float!')
    if null is not None:
        data = copy.deepcopy(data)
        if isinstance(data, gs.DataFile):
            data.data.fillna(value=null, inplace=True)
        elif isinstance(data, pd.DataFrame):
            data.fillna(value=null, inplace=True)
        else:
            # Shouldn't get here... but just in case
            raise ValueError('only DataFrame and DataFile is implemented!')
    return data


def write_hvtk(data, flname, griddef, variables=None):
    """
    Writes out an H5 file and corresponding xdmf file that Paraview can read. Currently only
    supports 3D gridded datasets. This function will fail if the length of the DataFile or
    DataFrame does not equal ``griddef.count()``.

    The extension xdmf is silently enforced. Any other extension passed is replaced.

    Parameters:
        data (pd.DataFrame): The DataFrame to writeout
        flname (str): Path (or name) of file to write out.
        griddef (GridDef): Grid definitions for the realizations to be written out
        variables (str or list): optional set of variables to write out from the DataFrame

    """
    import os
    import subprocess
    from .h5_io import write_h5
    # setup the temporary strings to write things too:
    temp_attr = """           <Attribute Name="{varname}"
             AttributeType="Scalar"
             Center="Cell">
           <DataItem Dimensions="{nx} {ny} {nz}" NumberType="Float" Precision="8" Format="HDF">
            {h5flname}:{attrvarpath}
           </DataItem>
         </Attribute>"""
    xmlstr = """<?xml version="1.0" ?>
    <!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
    <Xdmf Version="2.0">
     <Domain>
       <Grid Name="mesh" GridType="Uniform">
         <Topology TopologyType="3DCoRectMesh" Dimensions="{nz} {ny} {nx}"></Topology>
          <Geometry Type="ORIGIN_DXDYDZ">
            <DataItem Format="XML" Dimensions="{dim}">{zmn} {ymn} {xmn}</DataItem>
            <DataItem Format="XML" Dimensions="{dim}">{zsiz}  {ysiz}  {xsiz}</DataItem>
          </Geometry>
{attributes}
      </Grid>
     </Domain>
    </Xdmf>
    """
    # Checks
    if not isinstance(data, pd.DataFrame):
        print('ERROR: Pass a pd.DataFrame for writeout!')
        return
    if 1 in (griddef.nx, griddef.ny, griddef.nz) or len(data) != griddef.count():
        print('ERROR: This function only supports 3D grids! Check data or griddefs!')
        return

    attributes = ''
    ext = flname[flname.rfind('.'):]
    if ext.lower() != '.xdmf':
        flname = flname.replace(ext, '.xdmf')
        print('WARNING: Paraview h5-reader likes xdmf extensions - \nnew file'
              ' is: %s' % flname)
    outh5name = flname.replace('.xdmf', '.hvtk')

    # Always overwrites the old hvtk file, but has to clear it first
    if os.path.isfile(outh5name):
        try:
            os.remove(outh5name)
        except:
            try:
                subprocess.call('rm ' + outh5name)
            except:
                print('ERROR: os.remove failed, and `rm` not found on this system. Please '
                      'remove %s manually' % outh5name)
                return

    # do some string enchantments
    if '/' in outh5name or '\\' in outh5name:
        h5path = outh5name
        # take `\` and `/` out of the path
        while '/' in h5path:
            h5path = h5path[h5path.rfind('/') + 1:]
        while '\\' in h5path:
            h5path = h5path[h5path.rfind('\\') + 1:]
    else:
        h5path = outh5name
    h5path = h5path.replace(' ', '')

    # write out all the data in the dataframe if no variables are passed
    if variables is not None:
        if isinstance(variables, str):
            variables = [variables]
        data = data[variables]

    # write out each variable to the hvtk data storage
    for var in data.columns:
        variable = var.lower().replace(' ', '')
        varpath = 'data/%s' % variable
        write_h5(data[var].values, outh5name, h5path=varpath)
        varpath += '/data'
        attributes += temp_attr.format(nx=griddef.nx, ny=griddef.ny, nz=griddef.nz,
                                       h5flname=h5path, varname=variable,
                                       attrvarpath=varpath) + '\n'

    # write out the XDMF pointing to the correct attributes in the hvtk
    xmlstr_nx = griddef.nx + 1
    xmlstr_ny = griddef.ny + 1
    xmlstr_nz = griddef.nz + 1
    xmlstr_xmn = griddef.xmn - (griddef.xsiz / 2)
    xmlstr_ymn = griddef.ymn - (griddef.ysiz / 2)
    xmlstr_zmn = griddef.zmn - (griddef.zsiz / 2)
    xmlstr_xsiz = griddef.xsiz
    xmlstr_ysiz = griddef.ysiz
    xmlstr_zsiz = griddef.zsiz

    xmlstr = xmlstr.format(attributes=attributes, nx=xmlstr_nx, ny=xmlstr_ny, nz=xmlstr_nz,
                           xmn=xmlstr_xmn, ymn=xmlstr_ymn, zmn=xmlstr_zmn, xsiz=xmlstr_xsiz,
                           ysiz=xmlstr_ysiz, zsiz=xmlstr_zsiz, dim=3)

    with open(flname, 'w') as fh:
        fh.write(xmlstr)


def file_nlines(flname):
    '''Open a file and get the total number of lines. Seems pretty fast. Copied from stackoverflow
    http://stackoverflow.com/questions/845058/how-to-get-line-count-cheaply-in-python

    Parameters:
        flname (str): Name of the file to read
    '''
    f = open(flname)
    lines = 0
    buf_size = 1024 * 1024
    read_f = f.read

    buf = read_f(buf_size)
    while buf:
        lines += buf.count('\n')
        buf = read_f(buf_size)
    return lines


def writeout_gslib_gmm(gmm, outfile):
    """
    Writeout a fitted Gaussian mixture to the format consistent with ``gmmfit`` from the CCG
    Knowledge Base. Assume ``gmm`` is a an ``sklearn.mixture.GaussianMixture`` class fitted to data

    Note:
        Recently GMM was replaced with GaussianMixture, and there are subtle differences in
        attributes between the different versions..

    Parameters:
        gmm (GaussianMixture): a fitted mixture model
        outfile (str): the output file

    """
    if not gmm.converged_:
        print('GMM has not converged! ')
        return
    gmmstr = ''
    nvar = np.size(gmm.means_, 1)
    # write the header:
    gmmstr += 'sklearn.mixture.GMM Model for CCG gmm programs \n'
    gmmstr += '%i  %i \n' % (gmm.n_components, nvar)
    for i in range(gmm.n_components):
        # weights on one line
        gmmstr += '%.16g  %.16g \n' % (0.0, gmm.weights_[i])
        # means on one line
        gmmstr += ' '.join(['%.16g' % val for val in gmm.means_[i, :]])
        gmmstr += '\n'
        # covariance on another line
        for j in range(nvar):
            for k in range(j, nvar):
                gmmstr += '%.16g ' % gmm.covariances_[i, j, k]
        gmmstr += '\n'
    with open(outfile, 'w') as file:
        file.write(gmmstr)
    file.close()


def readvarg(vargflname, vargnum):
    """
    Reads in a variogram file and returns a pygeostat data file or list of pygeostat DataFiles with
    the variograms

    The parameter `vargnum` can be one of the following:

        * a single variogram number to return one pygeostat DataFile
        * a list of variogram numbers to return a list of DataFiles
        * the string 'all' to return a list with all variograms

    The returned `gs.DataFrame` objects contain the following parameters:

        * data['Number'] - lag number
        * data['Distance'] - lag distance
        * data['Value'] - variogram value
        * data['Points'] - number of points
        * data['Head'] - head value of variogram
        * data['Tail'] - tail value of variogram
        * tailname - tail variable name
        * headname - head variable name
        * direction - direction number

    or for gamv2004 variograms:

        * azm - azimuth
        * azmtol - azimuth tolerance
        * azmbandw - azimuth bandwidth
        * dip - dip
        * diptol - dip tolerance
        * dipbandw - dip bandwidth
        * lags - number of lags
        * lagdist - lag distance
        * lagtol - lag tolerance
        * vartype - variogram type
        * tailvar - tail variable number
        * headvar - head variable number
        * indcatcut - indicator category or cutoff (None if missing)

    Parameters:
        vargflname: name of variogram file in either gamv or gamv2004 style
        vargnum: permissible vargnum parameter as specified above

    Returns:
        vargs: A gs.DataFile (or list of DataFiles) with data set to the variogram data as
            as specified above
    """
    import io
    from . data import DataFile

    # What variogram number/numbers are we reading:
    if isinstance(vargnum, str):
        if vargnum.lower() == 'all':
            # Count the number of variograms
            vargnum = 1
            with open(vargflname, 'r') as vargfl:
                for line in vargfl:
                    if line[:1].isalpha():
                        vargnum += 1
            vargnum = list(range(1, vargnum))
        else:
            raise Exception('Invalid vargnum', vargnum)
    # Convert vargnum to list if just reading a single variogram
    if isinstance(vargnum, int):
        vargnum = [vargnum]
    # Read in each variogram
    vargidx = 0
    vargs = []
    with open(vargflname, 'r') as vargfl:
        savedline = None
        for line in vargfl:
            if not (line[:1].isalpha()) and savedline is None:
                continue
            # Increment the variogram index number
            vargidx += 1
            # Are we reading this variogram?
            if vargidx not in vargnum:
                savedline = None
                continue
            varg = DataFile(flname=vargflname, readfl=False, fltype='gslib')
            varg.vartype = '-1'
            # Get the head variable, tail variable and direction
            if savedline is None:
                headerline = line
            else:
                headerline = savedline
            # Try and get header data, might not work for vmodel outputs
            try:
                varg.tailname = headerline.partition('tail:')[2].partition('head:')[0].strip()
                if 'direction:' in headerline:
                    varg.headname = headerline.partition('head:')[2].partition('direction:')[0].strip()
                else:
                    varg.headname = headerline.partition('head:')[2].partition('direction')[0].strip()
                varg.direction = int(headerline.partition('direction')[2])
            except ValueError:
                pass
            if savedline is None:
                line = next(vargfl)
            # Check for gamv2004 output type
            if line.startswith('-HDIR'):
                # Read in gamv2004 specific information
                varg.azm = line.split()[1]
                varg.azmtol = line.split()[2]
                varg.azmbandw = line.split()[3]
                line = next(vargfl)
                varg.dip = line.split()[1]
                varg.diptol = line.split()[2]
                varg.dipbandw = line.split()[3]
                line = next(vargfl)
                varg.lags = line.split()[1]
                varg.lagdist = line.split()[2]
                varg.lagtol = line.split()[3]
                line = next(vargfl)
                varg.vartype = line.split()[1]
                varg.tailvar = line.split()[2]
                varg.headvar = line.split()[3]
                try:
                    varg.indcatcut = line.split()[3]
                except IndexError:
                    varg.indcatcut = None
                vargstr = next(vargfl)
            else:
                vargstr = line
            while True:
                line = None
                # Try and read the next line
                try:
                    line = next(vargfl)
                except StopIteration:
                    break
                # Add this line to the variogram string
                if line is not None:
                    if not line[:1].isalpha():
                        vargstr = vargstr + line
                    else:
                        savedline = line
                        break
            iovarg = io.StringIO(vargstr)
            # Read in the data
            if varg.vartype.strip() != '4':
                varg.data = pd.read_csv(iovarg, header=None, delim_whitespace=True,
                                        skipinitialspace=True,
                                        names=['Number', 'Distance', 'Value', 'Points',
                                               'Head', 'Tail'],
                                        engine='python')
            else:
                varg.data = pd.read_csv(iovarg, header=None, delim_whitespace=True,
                                        skipinitialspace=True,
                                        names=['Number', 'Distance', 'Value', 'Points',
                                               'Head', 'Tail', 'HeadVar', 'TailVar'],
                                        engine='python')
            vargs.append(varg)
    # Return with variogram(s)
    if len(vargs) == 1:
        return vargs[0]
    else:
        return vargs


'''
Pygeostat plotting modules
'''
from __future__ import absolute_import, division, print_function

# Load some plotting modules
from .histogram_plot import histogram_plot
from .histogram_plot_simulation import histogram_plot_simulation
from .slice_plot import slice_plot
from .scatter_plot import *
from .pit_plot import pit_plot
from .export_image import export_image
from .loadings_plot import loadings_plot
from .validation_plot import validation_plot
from .correlation_matrix_plot import correlation_matrix_plot
from .variogram_plot import  variogram_plot
from .variogram_plot_simulation import variogram_plot_simulation
from .location_plot import location_plot
from .grid_slice_plot import grid_slice_plot
from .image_grid import image_grid
from .set_style import *
from .variogram_utils import *
from .cmaps import *
from .utils import *
from .accuracy_plot import accuracy_plot
from .drill_plot import drill_plot
from .mds_plot import mds_plot
from .probability_plot import probability_plot
from .qq_plot import qq_plot
from .subplots import *
from .contour_plot import contour_plot
